from functools import partial

from PySide6.QtWidgets import (
    QApplication, QVBoxLayout, QHBoxLayout, QWidget,
    QLabel, QComboBox, QFrame, QPushButton, QGroupBox, QMessageBox,
    QSizePolicy, QSplitter, QSpinBox, QScrollArea,
    QGraphicsDropShadowEffect, QStackedLayout
)
from PySide6.QtCore import Qt, QModelIndex, QSize, Signal, QThreadPool, QTimer
from PySide6.QtGui import QColor

import matplotlib as mpl
from matplotlib import collections
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvas

from combo_selector.ui.widgets.custom_toolbar import CustomToolbar
from combo_selector.core.plot_utils import PlotUtils
from combo_selector.core.orthogonality import Orthogonality
from combo_selector.core.workers import OMWorkerComputeOM, OMWorkerUpdateNumBin
from combo_selector.ui.widgets.line_widget import LineWidget
from combo_selector.ui.widgets.circle_progress_bar import RoundProgressBar
from combo_selector.ui.widgets.style_table import StyledTable
from combo_selector.ui.widgets.checkable_tree_list import CheckableTreeList

PLOT_SIZE = QSize(600, 400)

METRIC_PLOT_MAP = {
    "Convex hull relative area": "Convex Hull",
    "Bin box counting": "Bin Box",
    "Pearson Correlation": "Linear regression",
    "Spearman Correlation": "Linear regression",
    "Kendall Correlation": "Linear regression",
    "Asterisk equations": "Asterisk",
    "%FIT": "%FIT yx",
    "%BIN": "%BIN",
    "Gilar-Watson method": None,
    "Modeling approach": "Modeling approach",
    "Geometric approach": "Geometric approach",
    "Conditional entropy": "Conditional entropy",
    "NND Arithm mean": None,
    "NND Geom mean": None,
    "NND Harm mean": None,
    "NND mean": None
}

class OMCalculationPage(QFrame):

    metric_computed = Signal(list)
    gui_update_requested = Signal()

    def __init__(self, model: Orthogonality = None):
        super().__init__()
        self.selected_metric_list = []
        self.threadpool = QThreadPool()
        self.selected_scatter_collection = None
        self.selected_metric = None  # First metric in the list
        self.artist_list = []
        self.arrow_list = []
        self.selected_set = "Set 1"
        self.orthogonality_dict = {}
        self.model = model

        self.fig = Figure(figsize=(15,15))
        self.canvas = FigureCanvas(self.fig)
        self.toolbar = CustomToolbar(self.canvas)

        self.selected_axe = self.canvas.figure.add_subplot(1, 1, 1)
        self.selected_axe.set_box_aspect(1)
        self.selected_axe.set_xlim(0, 1)
        self.selected_axe.set_ylim(0, 1)


        self.plot_utils = PlotUtils(fig=self.fig)

        self.plot_functions_map = {
            "Convex Hull": partial(self.plot_convex_hull),
            "Bin Box": partial(self.plot_bin_box),
            "Linear regression": partial(self.plot_linear_reg),
            "Asterisk": partial(self.plot_utils.plot_asterisk),
            "%FIT xy": partial(self.plot_utils.plot_percent_fit_xy),
            "%FIT yx": partial(self.plot_utils.plot_percent_fit_yx),
            "%BIN": partial(self.plot_utils.plot_percent_bin),
            "Modeling approach": partial(self.plot_utils.plot_modeling_approach),
            "Conditional entropy": partial(self.plot_utils.plot_conditional_entropy)
        }

        self.setFrameShape(QFrame.StyledPanel)
        self.setFrameShadow(QFrame.Raised)

        self.shadow = QGraphicsDropShadowEffect(self)
        self.shadow.setBlurRadius(20)
        self.shadow.setXOffset(0)
        self.shadow.setYOffset(0)
        self.shadow.setColor(QColor(0, 0, 0, 100))

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        top_frame = QFrame()
        top_frame.setGraphicsEffect(self.shadow)
        top_frame_layout = QHBoxLayout(top_frame)
        top_frame_layout.setContentsMargins(25, 25, 25, 25)
        top_frame_layout.setSpacing(25)

        user_input_scroll_area = QScrollArea()

        input_title = QLabel("Input")
        input_title.setFixedHeight(30)
        input_title.setObjectName("TitleBar")
        input_title.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        input_title.setContentsMargins(10, 0, 0, 0)
        input_title.setStyleSheet("""
            background-color: #154E9D;
            color: white;
            font-weight:bold;
            font-size: 16px;
            border-top-left-radius: 12px;
            border-top-right-radius: 12px;
        """)

        user_input_scroll_area.setWidgetResizable(True)
        user_input_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        user_input_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        user_input_frame = QFrame()
        user_input_frame.setStyleSheet("background-color: white; border-radius: 10px;")
        user_input_frame_layout = QVBoxLayout(user_input_frame)
        user_input_frame_layout.setContentsMargins(20, 20, 20, 20)
        user_input_scroll_area.setWidget(user_input_frame)

        input_section = QFrame()
        input_section.setFixedWidth(290)
        input_layout = QVBoxLayout(input_section)
        input_layout.setSpacing(0)
        input_layout.setContentsMargins(0, 0, 0, 0)
        input_layout.addWidget(input_title)
        input_layout.addWidget(user_input_scroll_area)

        # OM Calculation Group
        om_computing_group = QGroupBox("OM calculation")
        om_calculation_layout = QVBoxLayout()
        om_calculation_layout.setContentsMargins(5, 5, 5, 5)
        om_computing_group.setLayout(om_calculation_layout)

        om_computing_group.setStyleSheet("""
             QGroupBox {
                font-size: 14px;
                font-weight: bold;
                background-color: #e7e7e7;
                color: #154E9D;
                border: 1px solid #d0d4da;
                border-radius: 12px;
                margin-top: 25px;
            }
            QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 0px;
                margin-top: -8px;
            }
                QPushButton {
                background-color: #d5dcf9;
                color: #2C3346;
                border: none;
                border-radius: 6px;
                padding: 8px 16px;
                font-weight: 500;
            }
            QPushButton:hover {
                background-color: #bcc8f5;
            }
            QPushButton:pressed {
                background-color: #8fa3ef;
            }
            QPushButton:disabled {
                background-color: #E5E9F5;
                color: #FFFFFF;
            }
                           QLabel {
            background-color: transparent;
            }
            
            
             QLabel#sub-title {
            background-color: transparent;
            color: #2C3E50;
            font-family: "Segoe UI";
            font-weight: bold;
            }
        """)

        metric_list = [
            "Convex hull relative area", "Bin box counting","Gilar-Watson method","Modeling approach","Conditional entropy",
            "Pearson Correlation", "Spearman Correlation", "Kendall Correlation", "Asterisk equations",
            "NND Arithm mean", "NND Geom mean", "NND Harm mean", "%FIT", "%BIN"
        ]

        self.om_tree_list = CheckableTreeList(metric_list)
        self.om_tree_list.setFixedHeight(175)

        self.om_calculate_btn = QPushButton("Compute metrics")
        self.footnote = QLabel()
        self.footnote.setTextFormat(Qt.TextFormat.RichText)
        self.footnote.setWordWrap(True)
        self.footnote.setText("<strong>NND</strong>: Nearest Neighbor Distance")
        self.footnote.setStyleSheet("font-size: 8pt;")

        number_of_bin_layout = QHBoxLayout()
        self.nb_bin = QSpinBox()
        self.nb_bin.setFixedWidth(100)
        self.nb_bin.setValue(14)
        self.nb_bin_label = QLabel('Number of bin box:')
        self.nb_bin_label.setObjectName("sub-title")
        number_of_bin_layout.addWidget(self.nb_bin_label)
        number_of_bin_layout.addWidget(self.nb_bin)

        select_metric_title = QLabel("Select metrics to compute:")
        select_metric_title.setObjectName("sub-title")
        om_calculation_layout.addWidget(select_metric_title)
        om_calculation_layout.addWidget(self.om_tree_list)
        om_calculation_layout.addWidget(self.footnote)
        # om_calculation_layout.addWidget(LineWidget('Horizontal'))
        om_calculation_layout.addSpacing(15)
        om_calculation_layout.addLayout(number_of_bin_layout)
        om_calculation_layout.addSpacing(15)
        om_calculation_layout.addWidget(self.om_calculate_btn)

        data_selection_group = QGroupBox("OM selection")
        data_selection_group.setStyleSheet("""
               QGroupBox {
                font-size: 14px;
                font-weight: bold;
                background-color: #e7e7e7;
                color: #154E9D;
                border: 1px solid #d0d4da;
                border-radius: 12px;
                margin-top: 25px;
            }
            QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 0px;
                margin-top: -8px;
            }
               QLabel {
            background-color: transparent;
            color: #2C3E50;
            font-family: "Segoe UI";
            font-weight: bold;
            }

            QComboBox::drop-down {
                border: none;
            }
            QComboBox::down-arrow {
                image: url(./drop_down_arrow.png);
            }
        """)

        om_selection_layout = QVBoxLayout()
        om_selection_layout.addWidget(QLabel("Number of metric to compare:"))
        self.compare_number = QComboBox()
        self.compare_number.addItems(["1", "2", "3", "4"])
        om_selection_layout.addWidget(self.compare_number)
        om_selection_layout.addSpacing(20)


        om_selection_layout.addWidget(QLabel("Select data set:"))
        self.dataset_selector = QComboBox()
        om_selection_layout.addWidget(self.dataset_selector)



        self.om_selector1 = QComboBox()
        self.om_selector2 = QComboBox()
        self.om_selector3 = QComboBox()
        self.om_selector4 = QComboBox()

        self.om_selector2.setDisabled(True)
        self.om_selector3.setDisabled(True)
        self.om_selector4.setDisabled(True)

        om_selection_layout.addWidget(QLabel("Select OM 1:"))
        om_selection_layout.addWidget(self.om_selector1)

        om_selection_layout.addWidget(QLabel("Select OM 2:"))
        om_selection_layout.addWidget(self.om_selector2)

        om_selection_layout.addWidget(QLabel("Select OM 3:"))
        om_selection_layout.addWidget(self.om_selector3)

        om_selection_layout.addWidget(QLabel("Select OM 4:"))
        om_selection_layout.addWidget(self.om_selector4)

        self.om_selector_list = [
            self.om_selector1,
            self.om_selector2,
            self.om_selector3,
            self.om_selector4
        ]

        self.om_selector_map = {
            '0': {'selector': self.om_selector1, 'axe': None, 'scatter_collection': None},
            '1': {'selector': self.om_selector2, 'axe': None, 'scatter_collection': None},
            '2': {'selector': self.om_selector3, 'axe': None, 'scatter_collection': None},
            '3': {'selector': self.om_selector4, 'axe': None, 'scatter_collection': None}
        }

        data_selection_group.setLayout(om_selection_layout)

        # Tips Section
        page_tips_group = QGroupBox("Tips")
        page_tips_group.setStyleSheet("""
            QGroupBox {
                font-size: 14px;
                font-weight: bold;
                background-color: #e7e7e7;
                color: #154E9D;
                border: 1px solid #d0d4da;
                border-radius: 12px;
                margin-top: 25px;
            }
            QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 0px;
                margin-top: -8px;
            }
        """)

        page_tips_layout = QVBoxLayout()
        self.textEdit = QLabel()
        self.textEdit.setTextFormat(Qt.TextFormat.RichText)
        self.textEdit.setWordWrap(True)
        page_tips_layout.addWidget(self.textEdit)
        page_tips_group.setLayout(page_tips_layout)

        user_input_frame_layout.addWidget(om_computing_group)
        user_input_frame_layout.addWidget(LineWidget("Horizontal"))
        user_input_frame_layout.addWidget(data_selection_group)
        # user_input_frame_layout.addWidget(LineWidget("Horizontal"))
        # user_input_frame_layout.addStretch()
        # user_input_frame_layout.addWidget(page_tips_group)

        # Plot Section
        plot_frame = QFrame()
        plot_frame = QFrame()
        plot_frame.setStyleSheet("""
            background-color: #e7e7e7;
            border-top-left-radius: 10px;
            border-top-right-radius: 10px;
        """)
        plot_frame_layout = QVBoxLayout(plot_frame)
        plot_frame_layout.setContentsMargins(0, 0, 0, 0)

        plot_title = QLabel("OM visualization")
        plot_title.setFixedHeight(30)
        plot_title.setObjectName("TitleBar")
        plot_title.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        plot_title.setContentsMargins(10, 0, 0, 0)
        plot_title.setStyleSheet("""
            background-color: #154E9D;
            color: white;
            font-weight:bold;
            font-size: 16px;
            font-weight: bold;
            padding: 6px 12px;
            border-top-left-radius: 10px;
            border-top-right-radius: 10px;
        """)




        plot_frame_layout.addWidget(plot_title)
        plot_frame_layout.addWidget(self.toolbar)
        plot_frame_layout.addWidget(self.canvas)

        # Combine input and plot
        top_frame_layout.addWidget(input_section)
        top_frame_layout.addWidget(plot_frame)

        # Table Section
        table_frame = QFrame()
        table_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Plain)

        table_frame_layout = QHBoxLayout(table_frame)
        table_frame_layout.setContentsMargins(20, 20, 20, 20)

        self.styled_table = StyledTable("OM result table")
        self.styled_table.set_header_label(["Set #", "2D Combination", "OM 1", "OM 2", "...", "OM n"])
        self.styled_table.set_default_row_count(10)

        table_frame_layout.addWidget(self.styled_table)

        # --- Progress Overlay Setup ---
        self.progress_bar = RoundProgressBar()
        self.progress_bar.rpb_setBarStyle('Pizza')

        self.progress_overlay = QWidget(self)
        self.progress_overlay.setAttribute(Qt.WA_TransparentForMouseEvents)
        self.progress_overlay.setStyleSheet("background-color: transparent;")
        self.progress_overlay.hide()

        # Overlay layout and progress bar placement
        overlay_layout = QVBoxLayout(self.progress_overlay)
        overlay_layout.setContentsMargins(0, 0, 0, 0)
        overlay_layout.addStretch()
        overlay_layout.addWidget(self.progress_bar, alignment=Qt.AlignCenter)
        overlay_layout.addStretch()

        # Main layout widget for content
        self.main_widget = QWidget()
        self.main_layout = QVBoxLayout(self.main_widget)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)

        # --- Main Splitter ---
        self.main_splitter = QSplitter(Qt.Vertical, self)
        self.main_splitter.addWidget(top_frame)
        self.main_splitter.addWidget(table_frame)
        self.main_layout.addWidget(self.main_splitter)

        # Stacked widget to hold content and overlay
        self.stack = QStackedLayout()
        self.stack.setStackingMode(QStackedLayout.StackingMode.StackAll)
        self.stack.addWidget(self.main_widget)
        self.stack.addWidget(self.progress_overlay)
        self.stack.setCurrentWidget(self.progress_overlay)  # default view

        # Base layout holds the stack
        self.base_layout = QVBoxLayout(self)
        self.base_layout.setContentsMargins(0, 0, 0, 0)
        self.base_layout.addLayout(self.stack)

        # Ensure overlay tracks resizing
        self.progress_overlay.setGeometry(self.stack.geometry())
        self.progress_overlay.raise_()

        # Connections
        # self.dataset_selector.currentTextChanged.connect(self.data_sets_change)
        self.compare_number.currentTextChanged.connect(self.update_om_selector_state)
        for index, data in self.om_selector_map.items():
            data["selector"].currentTextChanged.connect(lambda _, k=index: self.on_selector_changed(k))

        self.om_calculate_btn.clicked.connect(self.compute_orthogonality_metric)
        self.nb_bin.editingFinished.connect(self.update_bin_box_number)

        self.dataset_selector.currentTextChanged.connect(self.data_set_selection_changed_from_combobox)
        # Table selection event
        # self.styled_table.selectionChanged.connect(self.data_set_selection_changed_from_table)

    # Override resizeEvent to keep overlay in sync
    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.progress_overlay.setGeometry(self.stack.geometry())

    def start_om_computation(self, metric_list):
        worker = OMWorkerComputeOM(metric_list, self.model)

        worker.signals.progress.connect(self.handle_progress_update)
        worker.signals.finished.connect(self.handle_finished)

        self.threadpool.start(worker)

    def start_update_bin_number(self,nb_bin):
        checked_metric_list = self.om_tree_list.get_checked_items()

        #TODO this was to only compute the metric if any of these metric were selected
        # remove the threading if not necessary

        if any(metric in checked_metric_list \
               for metric in ["Bin box counting", "Modeling approach", "Gilar-Watson method"]):

            worker = OMWorkerUpdateNumBin(nb_bin,checked_metric_list, self.model)

            worker.signals.progress.connect(self.handle_progress_update)

            #TODO find a better way to compute the metric once bin number has been updated
            # this is commented because this call this will update table and trigger computed_metric signal
            # the issue is that when bin number is updated, I should find a way to compute metric if they have at least
            # been computed once otherwise it means I compute the metric without clicking on compute metric button
            # worker.signals.finished.connect(self.handle_finished)


            self.threadpool.start(worker)
        else:
            return

    def handle_progress_update(self, value: int):
        if self.om_tree_list.get_checked_items():

            if value == 0:
                self.progress_overlay.hide()
            else:
                self.stack.setCurrentWidget(self.progress_overlay)
                self.progress_overlay.show()
                self.progress_bar.rpb_setValue(value)
                self.progress_bar.repaint()

            if value == 100:
                self.progress_bar.repaint()

            QApplication.processEvents()
        else:
            return

    def handle_finished(self):
        if self.om_tree_list.get_checked_items():
            self.progress_bar.rpb_setValue(100)  # Final visual update
            self.progress_bar.repaint()
            QTimer.singleShot(800, self.hide_progress_overlay)

            self.update_orthogonality_table()
            self.data_sets_change()
            self.metric_computed.emit([self.om_tree_list.get_checked_items(), self.selected_metric_list])
        else:
            return

    def hide_progress_overlay(self):
        self.progress_overlay.hide()
        self.stack.setCurrentWidget(self.main_widget)

    def init_page(self):
        self.om_tree_list.unchecked_all()
        self.styled_table.clean_table()
        self.styled_table.set_header_label(["Set #", "2D Combination", "OM 1", "OM 2", "...", "OM n"])
        # self.styled_table.set_table_data(pd.DataFrame())
        self.plot_utils.set_orthogonality_data(self.model.get_orthogonality_dict())
        self.model.reset_om_status_computation_state()
        self.populate_selector()
        self.update_om_selector_state()

    def clear_axis_except_pathcollections(self,ax):
        """Clear all elements from the axis except PathCollection objects (e.g., scatter plots)."""

        # Remove all lines
        for line in ax.get_lines():
            line.remove()

        for text in ax.texts:
            text.remove()

        # Remove all QuadMesh objects (used in pcolormesh, heatmaps)
        for quadmesh in ax.findobj(mpl.collections.QuadMesh):
            quadmesh.remove()

        # Remove all artists except PathCollection (scatter plots)
        for artist in ax.get_children():
            if isinstance(artist, mpl.legend.Legend):
                artist.remove()  # Remove legend
            elif isinstance(artist, mpl.axis.Axis):
                artist.set_visible(False)  # Hide ticks/labels

        # Refresh the figure
        # ax.figure.canvas.draw_idle()

    def update_om_selector_state(self):

        #TODO number_of_selector should be a class attribute (when compare_number currentTextChanged)
        number_of_selectors = int(self.compare_number.currentText())


        [self.om_selector_list[i].setDisabled(False) if i<number_of_selectors
         else self.om_selector_list[i].setDisabled(True) for i,selector in enumerate(self.om_selector_list)]

        self.update_plot_layout()

        self.refresh_displayed_plot()



    def update_plot_layout(self):
        #get the number of plot to compare
        number_of_selectors = self.compare_number.currentText()

        #create a key string based on the compare number value in order to know which ploy layout to select
        plot_key = number_of_selectors+'PLOT'


        # plot layout map that contains the list of plot layout to display based on the compare number
        plot_layout_map = {'1PLOT':[111,None,None,None],
                       '2PLOT':[121,122,None,None],
                       '3PLOT':[221,222,223,None],
                       '4PLOT':[221,222,223,224]}

        #get list of layout
        layout_list = plot_layout_map[plot_key]

        self.fig.clear()
        # self.remove_all_axes()

        for i,layout in enumerate(layout_list):
            index = str(i)
            #initialize selector axe and scatter point selection
            if layout is not None:
                axe = self.fig.add_subplot(layout)
                self.fig.subplots_adjust(wspace=.5, hspace=.5)
                axe.set_box_aspect(1)
                axe.set_xlim(0, 1)
                axe.set_ylim(0, 1)

                self.draw_figure()

                self.om_selector_map[index]['axe'] = axe
                self.om_selector_map[index]['scatter_collection'] = axe.scatter([], [],s=20, c='k', marker='o', alpha=0.5)
            else:
                self.om_selector_map[index]['axe'] = None
                self.om_selector_map[index]['scatter_collection'] = None

        # self.update_figure()

    def on_selector_changed(self, index):
        """Handle combobox text change and get the corresponding axe."""
        selector = self.om_selector_map[index]["selector"]
        self.selected_metric = selector.currentText()
        self.plot_utils.set_axe(self.om_selector_map[index]["axe"])
        self.plot_utils.set_scatter_collection(self.om_selector_map[index]["scatter_collection"])

        self.update_figure()

    def populate_selector(self):
        """
        Updates the dataset selection combo box and figure list with the available data sets.

        - Retrieves the list of available data sets from `self.orthogonality_dict`.
        - Updates `self.set_combo` with the new dataset options.
        - Ensures signals are blocked during updates to prevent unwanted UI triggers.

        Notes:
        - This method should be called after loading new data.
        - If no data sets are available, the combo boxes remain unchanged.
        """
        self.orthogonality_dict = self.model.get_orthogonality_dict()
        if not self.orthogonality_dict:
            return  # No data available, prevent unnecessary UI updates

        data_sets_list = list(self.orthogonality_dict.keys())

        self.dataset_selector.blockSignals(True)  # Prevent UI signal loops

        self.dataset_selector.clear()

        self.dataset_selector.addItems(data_sets_list)

        self.dataset_selector.blockSignals(False)

        #clear selector list in case previous data were loaded
        for index, data in self.om_selector_map.items():
            om_selector = data['selector']

            om_selector.blockSignals(True)
            om_selector.clear()
            om_selector.blockSignals(False)

    def compute_orthogonality_metric(self) -> None:
        self.selected_metric_list = self.om_tree_list.get_checked_items()

        # self.model.compute_orthogonality_metric(self.selected_metric_list)
        self.stack.setCurrentWidget(self.progress_overlay)
        self.progress_overlay.show()
        self.progress_bar.rpb_setValue(0)
        self.progress_bar.repaint()
        QApplication.processEvents()

        self.start_om_computation(self.selected_metric_list)
        self.selected_metric_list = [METRIC_PLOT_MAP[metric] for metric in self.selected_metric_list]

        #remove None item
        self.selected_metric_list = [metric for metric in self.selected_metric_list if metric]

        #rermove doublon from list (when converting a list into dict it removes dupplicate keys)
        self.selected_metric_list = list(dict.fromkeys(self.selected_metric_list))

        for index, data in self.om_selector_map.items():
            om_selector = data['selector']

            om_selector.blockSignals(True)
            om_selector.clear()
            om_selector.addItems(self.selected_metric_list)
            om_selector.blockSignals(False)

        #
        # self.update_orthogonality_table()
        #
        # self.data_sets_change()

        # self.metric_computed.emit([self.om_tree_list.get_checked_items(),self.selected_metric_list])


    def _on_om_computed(self, original_metric_list: list):
        # now the model has filled its dataframes; we can continue
        # exactly the same as before, e.g.:

        # remap to plotting names
        plot_list = [METRIC_PLOT_MAP[m] for m in original_metric_list]
        plot_list = [m for m in plot_list if m]
        plot_list = list(dict.fromkeys(plot_list))

        for idx, data in self.om_selector_map.items():
            sel = data['selector']
            sel.blockSignals(True)
            sel.clear()
            sel.addItems(plot_list)
            sel.blockSignals(False)

        self.update_orthogonality_table()
        self.data_sets_change()
        self.metric_computed.emit([original_metric_list, plot_list])


    def update_bin_box_number(self):

        #simply update the number of bin and reset computed metric status for the metric that use bin number
        self.model.update_num_bins(self.nb_bin.value())

        # self.stack.setCurrentWidget(self.progress_overlay)
        # self.progress_overlay.show()
        # self.progress_bar.rpb_setValue(0)
        # self.progress_bar.repaint()
        # QApplication.processEvents()
        #
        # self.start_update_bin_number(self.nb_bin.value())
        #
        # self.refresh_displayed_plot()
        #
        # if self.selected_metric_list:
        #     self.metric_computed.emit([self.om_tree_list.get_checked_items(), self.selected_metric_list])

    def select_orthogonality(self):
        self.model.set_orthogonality_value('orthogonality_score')

    def update_orthogonality_table(self):
        data = self.model.get_orthogonality_metric_df()

        self.styled_table.set_header_label(list(data.columns))
        self.styled_table.async_set_table_data(data)
        self.styled_table.set_table_proxy()

    def data_sets_change(self, data_set: str = None) -> None:

        """
        Handles dataset selection from the dropdown menu and updates the UI accordingly.

        - Updates the current dataset based on the selection.
        - Highlights the corresponding row in the table view.
        - Refreshes the displayed figure to reflect the new dataset.

        Parameters:
            data_set (str, optional): The dataset name to switch to.
                                      If None, it is retrieved from `self.set_combo`.

        Notes:
        - If triggered by the dropdown, `data_set` is None, so the function retrieves the selected value.
        - Updates both the dropdown and table selection to stay synchronized.
        - Calls `update_figure()` to refresh the visualization.
        """
        if data_set is None:
            self.selected_set = self.dataset_selector.currentText()
            index_at_row = self.get_index_from()

            # Prevent unnecessary signal emissions while selecting the row
            self.styled_table.get_table_view().blockSignals(True)
            self.styled_table.get_table_view().selectionModel().blockSignals(True)
            self.styled_table.get_table_view().selectRow(index_at_row)
            self.styled_table.get_table_view().blockSignals(False)
            self.styled_table.get_table_view().selectionModel().blockSignals(False)
        else:
            self.selected_set = data_set
            self.dataset_selector.blockSignals(True)
            self.dataset_selector.setCurrentText(self.selected_set)
            self.dataset_selector.blockSignals(False)

        self.plot_utils.set_set_number(self.selected_set)

        self.refresh_displayed_plot()

    def get_index_from(self) -> int:
        """
        Retrieves the row index corresponding to the currently selected dataset.

        - Iterates through the table model to find a match with `self.selected_set`.
        - Uses the proxy model to access the displayed data.

        Returns:g
            int: The row index of the selected dataset, or -1 if not found.

        Notes:
        - This function assumes that dataset names are stored in the first column.
        - If no match is found, returns -1 to indicate failure.
        """
        row_count = self.styled_table.get_row_count()

        for row in range(row_count):
            model_index = self.styled_table.get_proxy_model().index(row, 0)  # Column 0 assumed to contain dataset names
            if f"Set {model_index.data()}" == self.selected_set:
                return row

        return -1  # Explicitly return -1 if no match is found

    def table_item_clicked(self, index: QModelIndex) -> None:
        """
        Handles table row selection and updates the dataset accordingly.

        - Retrieves the dataset name from the selected row.
        - Calls `data_sets_change()` to synchronize the dropdown selection.

        Parameters:
            index (QModelIndex): The index of the clicked item (not used in current logic).

        Notes:
        - Only the first selected row is considered if multiple rows are selected.
        - Assumes dataset names are stored in the first column.
        """
        model_index_list = self.styled_table.get_selected_rows()

        if not model_index_list:
            return  # No selection, exit early

        self.data_sets_change(f"Set {model_index_list[0].data()}")

    def data_set_selection_changed_from_combobox(self) -> None:
        """
        Handles dataset selection changes from the combo box.

        - Updates the current dataset name.
        - Selects the corresponding row in the table view.
        - Refreshes the figure to display the selected dataset.

        Notes:
        - This function ensures synchronization between the combo box selection and the table view.
        - Calls `get_index_from()` to retrieve the correct row index.
        - If no matching row is found, no row selection is applied.
        """
        self.selected_set = self.dataset_selector.currentText()
        index_at_row = self.get_index_from()

        if index_at_row != -1:
            # signal is triggered by QItemSelectionModel not the QTableView when selecting a row
            self.styled_table.get_table_view().selectionModel().blockSignals(True)
            self.styled_table.select_row(index_at_row)
            self.styled_table.get_table_view().selectionModel().blockSignals(True)

        self.plot_utils.set_set_number(self.selected_set)

        self.refresh_displayed_plot()

    def data_set_selection_changed_from_table(self) -> None:
        """
        Handles dataset selection changes from the table view.

        - Retrieves the selected dataset from the table view.
        - Synchronizes the combo box with the selected dataset.
        - Updates the visualization based on the current tab.

        Notes:
        - Only the first selected row is considered if multiple rows are selected.
        - The combo box is blocked during updates to prevent signal loops.
        - If the "Calculations" tab is active, the selected dataset blinks on the orthogonality plot.
        - Otherwise, the standard figure update is triggered.
        """
        model_index_list = self.styled_table.get_selected_rows()

        if not model_index_list:
            return  # No selection, exit early

        proxy_model = self.styled_table.get_proxy_model()
        model_index = proxy_model.mapToSource(model_index_list[0])
        self.selected_set = f"Set {model_index.data()}"

        self.dataset_selector.blockSignals(True)
        self.dataset_selector.setCurrentText(self.selected_set)
        self.dataset_selector.blockSignals(False)

        self.refresh_displayed_plot()

    def refresh_displayed_plot(self):
        # Refresh the displayed figure
        number_of_selectors = int(self.compare_number.currentText())

        [self.on_selector_changed(str(i)) for i in range(number_of_selectors)]

    def draw_figure(self):
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

    def  update_figure(self):

        #if no axe is selected no need to update figure
        if self.selected_axe is None:
            return

        self.plot_utils.clean_figure()

        if self.model.get_status() in ['loaded', 'peak_capacity_loaded']:
            # self.plot_scatter()
            self.plot_utils.plot_scatter()
        else:
            return

        # Execute the appropriate method based on combo box selection
        # current_metric = self.computed_metric.currentText()

        # if self.selected_metric in ['Bin Box']:
        #     self.bin_box_input_widget.setHidden(False)
        # else:
        #     self.bin_box_input_widget.setHidden(True)

        #No metric has been computed yet so no need to plot them
        if self.selected_metric is None:
            return

        if self.selected_metric in self.plot_functions_map:
            self.plot_functions_map[self.selected_metric]()  # Call the corresponding function

    def plot_percent_bin(self):
        self.plot_utils.plot_percent_bin()

    def plot_bin_box(self):
        self.plot_utils.plot_bin_box()

    # Plot methods
    def plot_asterisk(self):
        self.plot_utils.plot_asterisk()

    def plot_linear_reg(self):
        self.plot_utils.plot_linear_reg()

    def plot_percent_fit_xy(self):
        self.plot_utils.plot_percent_fit_xy()

    def plot_percent_fit_yx(self):
        self.plot_utils.plot_percent_fit_yx()

    def plot_convex_hull(self):
        self.plot_utils.plot_convex_hull()


import sys
from PySide6.QtWidgets import QVBoxLayout, QFrame, QWidget,QPushButton,\
    QApplication, QHBoxLayout, QDialog, QLabel,QGridLayout,QStackedWidget

from PySide6.QtGui import QFont
from PySide6.QtCore import Signal

stylesheet = """
	#full_menu_widget {
		background-color: #12295F;
		border-top-left-radius :10px;
        border-top-right-radius : 0px; 
        border-bottom-left-radius : 10px; 
        border-bottom-right-radius : 0px;
	}

	/* style for QPushButton */
	#full_menu_widget QPushButton {
		border:none;
		font-size: 14px;
		border-radius: 3px;
		text-align: left;
		padding: 8px 0 8px 15px;
		color: #ffffff;
		background-color:#12295F ;
	}

	#full_menu_widget QPushButton:hover {
		    background-color: #3A4B6E;  /* Slightly lighter on hover */
    color: #FFFFFF;}
    
    	#full_menu_widget QPushButton:checked {
		    background-color: #3A4B6E;  /* Slightly lighter on hover */
    color: #FFFFFF;
	}


	/* style for logo image */
	#logo_label_2 {
		padding: 5px;
		color: #fff;
	}

	/* style for APP title */
	#logo_label_3 {
		padding-right: 10px;
		color: #fff;
	}
	
    #side_bar_title QLabel{
		padding-right: 10px;
		color: w;
	}
	
	
"""

class SideBar(QWidget):

    button_clicked = Signal(int)

    def __init__(self,title = 'Unnamed'):
        super().__init__()

        self.setStyleSheet(stylesheet)
        main_layout = QGridLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)

        self.full_menu_widget = QFrame(self)
        self.full_menu_widget.setFixedWidth(170)
        self.full_menu_widget.setObjectName('full_menu_widget')
        self.full_menu_layout = QVBoxLayout(self.full_menu_widget)

        side_bar_title_layout = QHBoxLayout(self.full_menu_widget)
        self.side_bar_logo = QLabel()
        self.side_bar_logo.setProperty("isBold", True)
        self.side_bar_logo.setText("X")
        self.side_bar_logo.setFixedSize(20,20)
        self.side_bar_logo.setScaledContents(True)

        self.side_bar_title = QLabel()
        self.side_bar_title.setProperty("isBold", True)
        self.side_bar_title.setText(title)
        self.side_bar_title.setObjectName('side_bar_title')
        font = QFont()
        font.setPointSize(15)
        self.side_bar_title.setFont(font)

        side_bar_title_layout.addWidget(self.side_bar_logo)
        #side_bar_title_layout.addWidget(self.side_bar_title)

        self.side_bar_button_frame = QFrame()
        self.side_bar_button_frame.setObjectName('side_bar_button_frame')
        self.side_bar_buttons_layout = QVBoxLayout(self.side_bar_button_frame)


        self.full_menu_layout.addLayout(side_bar_title_layout)
        #side_bar_title_layout.setAlignment(Qt.AlignTop)
        self.full_menu_layout.addWidget(self.side_bar_button_frame)
        self.full_menu_layout.addStretch()


        self.content_widget = QWidget(self)
        content_widget_layout = QVBoxLayout(self.content_widget)
        content_widget_layout.setContentsMargins(0,0,0,0)
        content_widget_layout.setSpacing(0)

        top_bar_widget = QWidget()
        top_bar_layout = QHBoxLayout(top_bar_widget)

        self.hide_side_bar_btn = QPushButton('X')

        top_bar_layout.addWidget(self.hide_side_bar_btn)
        top_bar_layout.addStretch()

        self.content_qstack = QStackedWidget()

        #content_widget_layout.addWidget(top_bar_widget)
        content_widget_layout.addWidget(self.content_qstack)


        #main_layout.addWidget(self.icon_only_widget, 0, 0, 1, 1)
        main_layout.addWidget(self.full_menu_widget, 0, 1, 1, 1)
        main_layout.addWidget(self.content_widget, 0, 2, 1, 1)

    def add_side_bar_button(self,name, widget):
        button = QPushButton(name)
        button.setFixedHeight(100)
        #button.setIcon(icon)
        button.setCheckable(True)
        button.setAutoExclusive(True)
        self.side_bar_buttons_layout.addWidget(button)

        self.content_qstack.addWidget(widget)

        widget.setUpdatesEnabled(False)
        button.clicked.connect(lambda x : self.content_qstack.setCurrentWidget(widget))
        # button.clicked.connect(lambda x : self.button_clicked.emit(self.content_qstack.currentIndex()))
        widget.setUpdatesEnabled(True)


        return button

if __name__ == "__main__":
    app = QApplication(sys.argv)

    w = QDialog()

    w.resize(950, 600)

    layout_main = QVBoxLayout(w)
    #layout_main.setContentsMargins(5, 5, 5, 5)

    SideBar = SideBar(title="2D Combo Selector")
    SideBar.add_side_bar_button('INPUT DATA', QLabel('INPUT DATA'))
    SideBar.add_side_bar_button('DATA PLOTTING\nPAIRWISE', QLabel('DATA PLOTTING\nPAIRWISE'))
    SideBar.add_side_bar_button('CALCULATION', QLabel('CALCULATION'))
    SideBar.add_side_bar_button('RESULTS', QLabel('RESULTS'))
    SideBar.add_side_bar_button('EXPORT', QLabel('EXPORT'))
    layout_main.addWidget(SideBar, 1)

    w.show()
    app.exec()


from PySide6.QtWidgets import (
    QWidget, QVBoxLayout, QHBoxLayout,
    QLabel, QFrame, QHeaderView
)
from PySide6.QtCore import Qt, QModelIndex, Signal, QThreadPool

from combo_selector.core.workers import TableDataWorker
from combo_selector.ui.widgets.orthogonality_table import OrthogonalityTableView, OrthogonalityTableModel



class StyledTable(QWidget):
    selectionChanged = Signal()
    def __init__(self,title=''):
        super().__init__()

        self.threadpool = QThreadPool()
        outer = QVBoxLayout(self)
        outer.setAlignment(Qt.AlignCenter)

        card = QFrame()
        card.setObjectName("CardFrame")
        card.setLayout(QVBoxLayout())
        card.layout().setContentsMargins(0, 0, 0, 0)
        card.layout().setSpacing(0)

        # Title bar
        title = QLabel(title)
        title.setFixedHeight(30)
        title.setObjectName("TitleBar")
        title.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        title.setContentsMargins(10, 0, 0, 0)

        # Table
        self.model = OrthogonalityTableModel()
        self.table = OrthogonalityTableView(self,self.model)


        # Footer
        footer = QFrame()
        footer.setObjectName("Footer")
        footer.setLayout(QHBoxLayout())
        footer.layout().setContentsMargins(12, 6, 12, 6)
        footer.layout().setSpacing(8)

        # combo1 = QComboBox()
        # combo1.addItems(["2020", "2021", "2022"])
        # combo2 = QComboBox()
        # combo2.addItems([">20,000", ">30,000", "All"])
        # combo3 = QComboBox()
        # combo3.addItems(["All", "Positive", "Negative"])

        # footer.layout().addWidget(combo1)
        # footer.layout().addWidget(combo2)
        # footer.layout().addStretch()
        # footer.layout().addWidget(combo3)

        # Assemble
        card.layout().addWidget(title)
        card.layout().addWidget(self.table)
        card.layout().addWidget(footer)
        outer.addWidget(card)

        self._apply_styles()

        self.table.selectionModel().selectionChanged.connect(self.selection_changed)

    def clean_table(self):
        self.model.set_formated_data([])
        self.set_default_row_count(10)

    def selection_changed(self):
        self.selectionChanged.emit()

    def async_set_table_data(self, df):
        worker = TableDataWorker(df, self.model.get_header_label())
        worker.signals.finished.connect(self.handle_data)
        self.threadpool.start(worker)

    def handle_data(self,data, rows, cols):
        self.model.apply_formatted_data(data, rows, cols)



    def set_table_data(self,data):
        self.model.set_data(data)
        # self.table.resizeColumnsToContents()

        for col in range(self.model.columnCount(QModelIndex())):
            current_width = self.table.columnWidth(col)
            self.table.setColumnWidth(col, current_width + 10)  # Add padding

        self.table.horizontalHeader().setDefaultAlignment(Qt.AlignLeft | Qt.AlignVCenter)

        self.table.horizontalHeader().setStretchLastSection(True)

        for i in range(self.table.model().columnCount(QModelIndex())):
            if i == 1:  # 'Combination' column index
                self.table.horizontalHeader().setSectionResizeMode(i, QHeaderView.Stretch)
            else:
                self.table.horizontalHeader().setSectionResizeMode(i, QHeaderView.ResizeToContents)

    def set_table_proxy(self):
        self.model.set_proxy(self.table.getProxyModel())

    def get_proxy_model(self):
        return self.table.getProxyModel()

    def get_selected_rows(self):
        return self.table.selectionModel().selectedRows()

    def get_model(self):
        return self.model

    def get_table_view(self):
        return self.table

    def get_row_count(self):
        return self.model.rowCount(QModelIndex())

    def select_row(self,index):
        self.table.selectRow(index)

    def set_header_label(self,header_label):
        self.model.set_header_label(header_label)

    def set_default_row_count(self,value):
        self.model.set_default_row_count(value)


    def _apply_styles(self):
        self.setStyleSheet("""
            QWidget {
                font-family: Segoe UI, Arial;
                font-size: 13px;
            }

            QFrame#CardFrame {
                background: white;
                border-radius: 12px;
                border: 1px solid #d0d5dd;
            }

            QLabel#TitleBar {
                background-color: #154E9D;
                color: white;
                font-weight:bold;
                font-size: 16px;
                border-top-left-radius: 12px;
                border-top-right-radius: 12px;
            }

    QHeaderView::section {
        background-color: #D1D9FC;
        color: #154E9D;
        font-size: 12px;
        padding: 4px;
        font-weight: bold;
        border: 1px solid #d0d4da;
    }

    QTableView {
    
        background-color: #F6F8FD;
        border: 1px solid #F6F8FD;
        gridline-color: #D4D6EC;
        selection-background-color: #c9daf8;
        selection-color: #000000;
        font-size: 11px;
    }

    QTableView::item {
        padding: 6px;Y
        padding: 2px;
        border: none;
    }

    QTableView::item:selected {
        background-color: #d8e5fc;
        color: #000000;
    }

    QScrollBar:vertical {
        border: none;
        background: #bdcaf6;
        width: 10px;
        margin: 4px 0 4px 0;
    }

    QScrollBar::handle:vertical {
        background: white;
        min-height: 20px;
        border-radius: 5px;
    }

    QScrollBar::add-line:vertical,
    QScrollBar::sub-line:vertical {
        height: 0;
    }

    QScrollBar::add-page:vertical,
    QScrollBar::sub-page:vertical {
        background: none;
    }

    QComboBox {
        background-color: #ffffff;
        border: 1px solid #c5d0e6;
        border-radius: 6px;
        padding: 5px 8px;
        font-size: 14px;
    }

            QFrame#Footer {
                background-color: #f1f3f6;
                border-bottom-left-radius: 12px;
                border-bottom-right-radius: 12px;
            }

            QComboBox {
                background-color: white;
                border: 1px solid #ccc;
                padding: 4px 8px;
                border-radius: 4px;
            }
        """)


import os

import matplotlib
matplotlib.use('Agg')  # make sure we’re using the non-GUI “Agg” backend

from functools import partial
import pandas as pd
from PySide6.QtGui import QColor
from PySide6.QtWidgets import (
    QVBoxLayout, QHBoxLayout, QFileDialog,
    QLabel, QLineEdit, QFrame, QPushButton, QGroupBox,
    QSizePolicy, QSplitter, QScrollArea,
    QGraphicsDropShadowEffect
)
from PySide6.QtCore import Qt, QTimer, QSize
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvas

from combo_selector.core.plot_utils import PlotUtils
from combo_selector.ui.widgets.custom_toolbar import CustomToolbar
from combo_selector.ui.widgets.checkable_tree_list import CheckableTreeList
from combo_selector.ui.widgets.checkable_combo_list import CheckableComboList
from combo_selector.ui.widgets.style_table import StyledTable
from combo_selector.ui.widgets.line_widget import LineWidget



PLOT_SIZE = QSize(600, 400)


class ExportPage(QFrame):
    def __init__(self, model=None, title='Unnamed'):
        super().__init__()

        self.blink_timer = QTimer()
        self.blink_step = 0
        self.blink_ax = None
        self.animations = []
        self.highlighted_ax = None
        self.selected_scatter_collection = None
        self.selected_axe = None
        self.full_scatter_collection = None
        self.selected_set = 'Set 1'
        self.orthogonality_dict = None
        self.model = model

        self.fig = Figure(figsize=(15,15))
        self.canvas = FigureCanvas(self.fig)
        self.toolbar = CustomToolbar(self.canvas)

        self.axe = self.canvas.figure.add_subplot(1, 1, 1)
        self.axe.set_box_aspect(1)
        self.axe.set_xlim(0, 1)
        self.axe.set_ylim(0, 1)


        self.plot_utils = PlotUtils(fig=self.fig)
        self.plot_utils.set_axe(self.axe)

        self.plot_functions_map = {
            "Convex Hull": partial(self.plot_convex_hull),
            "Bin Box": partial(self.plot_bin_box),
            "Linear regression": partial(self.plot_linear_reg),
            "Modeling approach": partial(self.plot_modeling_approach),
            "Conditional entropy": partial(self.plot_conditional_entropy),
            "Asterisk": partial(self.plot_utils.plot_asterisk),
            "%FIT": partial(self.plot_utils.plot_percent_fit_xy),
            "%BIN": partial(self.plot_utils.plot_percent_bin)
        }

        self.table_functions_map = {
            "Normalized retention table":self.model.get_normalized_retention_time_df,
            "2D Combination table":self.model.get_combination_df,
            "OM result table":self.model.get_orthogonality_metric_df,
             "Orthogonality result correlation table":self.model.get_correlation_group_df,
            "Final result and ranking table":self.model.get_orthogonality_result_df
        }

        self.shadow = QGraphicsDropShadowEffect(self)
        self.shadow.setBlurRadius(20)
        self.shadow.setXOffset(0)
        self.shadow.setYOffset(0)
        self.shadow.setColor(QColor(0, 0, 0, 100))

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.main_layout = QVBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)

        self.setFrameShape(QFrame.StyledPanel)
        self.setFrameShadow(QFrame.Raised)

        top_frame = QFrame()
        top_frame.setFrameShape(QFrame.NoFrame)
        top_frame.setGraphicsEffect(self.shadow)
        # f5f7fa

        top_frame_layout = QHBoxLayout(top_frame)
        top_frame_layout.setContentsMargins(25, 25, 25, 25)
        top_frame_layout.setSpacing(25)

        user_input_scroll_area = QScrollArea()
        user_input_scroll_area.setFixedWidth(290)
        input_title = QLabel("Input")
        input_title.setFixedHeight(30)
        input_title.setObjectName("TitleBar")
        input_title.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        input_title.setContentsMargins(10, 0, 0, 0)
        input_title.setStyleSheet("""
            background-color: #154E9D;
            color: white;
            font-weight:bold;
            font-size: 16px;
            border-top-left-radius: 12px;
            border-top-right-radius: 12px;
        """)

        user_input_scroll_area = QScrollArea()
        user_input_scroll_area.setFixedWidth(290)
        user_input_scroll_area.setWidgetResizable(True)
        user_input_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        user_input_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        user_input_frame = QFrame()
        user_input_frame.setStyleSheet("background-color: white; border-radius: 10px;")
        user_input_frame.setFixedWidth(290)
        user_input_frame_layout = QVBoxLayout(user_input_frame)
        user_input_frame_layout.setContentsMargins(20, 20, 20, 20)

        user_input_scroll_area.setWidget(user_input_frame)

        input_section = QFrame()
        input_layout = QVBoxLayout(input_section)
        input_layout.setSpacing(0)
        input_layout.setContentsMargins(0, 0, 0, 0)
        input_layout.addWidget(input_title)
        input_layout.addWidget(user_input_scroll_area)

        info_group = QGroupBox("Info")
        info_group.setStyleSheet("""
            QGroupBox {
                font-size: 14px;
                font-weight: bold;
                background-color: #e7e7e7;
                color: #154E9D;
                border: 1px solid #d0d4da;
                border-radius: 12px;
                margin-top: 25px;
            }
            QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 0px;
            }
            QLabel {
                background-color: transparent;
                color: #3f4c5a;
            }
        """)

        info_layout = QVBoxLayout()
        info_layout.addWidget(QLabel("Number of conditions:"))
        self.condition_label = QLabel("---")
        info_layout.addWidget(self.condition_label)
        info_layout.addWidget(QLabel("Number of combinations:"))
        self.combination_label = QLabel("---")
        info_layout.addWidget(self.combination_label)
        info_group.setLayout(info_layout)

        # Create the "Export data set figure" group box
        export_figure_grp = QGroupBox('Export data set figure')

        export_figure_grp.setStyleSheet("""
             QGroupBox {
                font-size: 14px;
                font-weight: bold;
                background-color: #e7e7e7;
                color: #154E9D;
                border: 1px solid #d0d4da;
                border-radius: 12px;
                margin-top: 25px;
            }
            QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 0px;
                margin-top: -8px;
            }
                QPushButton {
                background-color: #d5dcf9;
                color: #2C3346;
                border: none;
                border-radius: 6px;
                padding: 8px 16px;
                font-weight: 500;
            }
            QPushButton:hover {
                background-color: #bcc8f5;
            }
            QPushButton:pressed {
                background-color: #8fa3ef;
            }
            QPushButton:disabled {
                background-color: #E5E9F5;
                color: #FFFFFF;
            }
               QLabel {
            background-color: transparent;
            color: #2C3E50;
            font-family: "Segoe UI";
            font-weight: bold;
            }
            
            QComboBox::drop-down {
            border: none;
            }
            QComboBox::down-arrow {
                image: url(./drop_down_arrow.png);
            }
        """)




        # Use QFormLayout for aligning combo boxes and line edits
        form_layout = QVBoxLayout()

        # Export directory selection
        self.figure_export_directory_lineEdit = QLineEdit()
        self.figure_export_directory_lineEdit.setText(os.getcwd())
        # self.figure_export_directory_lineEdit.setFixedWidth(150)
        self.export_figure_directory_btn = QPushButton('...')
        self.export_figure_directory_btn.setFixedWidth(50)


        # Create a horizontal layout for the export directory section
        export_directory_hlayout = QHBoxLayout()
        export_directory_hlayout.addWidget(self.figure_export_directory_lineEdit)
        export_directory_hlayout.addWidget(self.export_figure_directory_btn)
        form_layout.addWidget(QLabel("Export directory:"))
        form_layout.addLayout(export_directory_hlayout)

        # Folder name input
        self.figure_folder_name_lineEdit = QLineEdit('Figure')
        # self.figure_folder_name_lineEdit.setFixedWidth(200)
        form_layout.addWidget(QLabel("Folder name:"))
        form_layout.addWidget(self.figure_folder_name_lineEdit)



        # Figure type selection
        self.figure_type_chklist = CheckableComboList()
        # self.figure_type_chklist.setFixedWidth(200)
        form_layout.addWidget(QLabel("Figure type:"))
        form_layout.addWidget(self.figure_type_chklist)

        # Figure list selection
        self.figure_list_chklist = CheckableComboList()
        # self.figure_list_chklist.setFixedWidth(200)
        form_layout.addWidget(QLabel("Figure list:"))
        form_layout.addWidget(self.figure_list_chklist)


        self.save_figure_btn = QPushButton('Save figure(s)')
        form_layout.addWidget(self.save_figure_btn)

        # Add the export directory section to the form layout

        export_figure_grp.setLayout(form_layout)



        # Create the "Export data set figure" group box
        export_table_grp = QGroupBox('Export table(s)')
        export_table_layout = QVBoxLayout()
        export_table_grp.setLayout(export_table_layout)

        export_table_grp.setStyleSheet("""
             QGroupBox {
                font-size: 14px;
                font-weight: bold;
                background-color: #e7e7e7;
                color: #154E9D;
                border: 1px solid #d0d4da;
                border-radius: 12px;
                margin-top: 25px;
            }
            QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 0px;
                margin-top: -8px;
            }
                QPushButton {
                background-color: #d5dcf9;
                color: #2C3346;
                border: none;
                border-radius: 6px;
                padding: 8px 16px;
                font-weight: 500;
            }
            QPushButton:hover {
                background-color: #bcc8f5;
            }
            QPushButton:pressed {
                background-color: #8fa3ef;
            }
            QPushButton:disabled {
                background-color: #E5E9F5;
                color: #FFFFFF;
            }
            QLabel {
            background-color: transparent;
            color: #2C3E50;
            font-family: "Segoe UI";
            font-weight: bold;
            }

            QComboBox::drop-down {
            border: none;
            }
            QComboBox::down-arrow {
                image: url(./drop_down_arrow.png);
            }
        """)

        table_list = ["Normalized retention table","2D Combination table","OM result table",
                      "Orthogonality result correlation table","Final result and ranking table"]

        self.table_selection = CheckableTreeList(table_list)
        self.table_selection.setFixedHeight(175)

        # Export directory selection
        self.table_export_directory_lineEdit = QLineEdit()
        self.table_export_directory_lineEdit.setText(os.getcwd())
        # self.figure_export_directory_lineEdit.setFixedWidth(150)
        self.export_table_directory_btn = QPushButton('...')
        self.export_table_directory_btn.setFixedWidth(50)

        self.export_filename = QLineEdit()
        self.export_filename.setText("export_table.xlsx")


        # Create a horizontal layout for the export directory section
        export_table_directory_hlayout = QHBoxLayout()
        export_table_directory_hlayout.addWidget(self.table_export_directory_lineEdit)
        export_table_directory_hlayout.addWidget(self.export_table_directory_btn)

        self.export_table_btn = QPushButton("Export table(s)")

        export_table_layout.addWidget(QLabel("Select table to export:"))
        export_table_layout.addWidget(self.table_selection)
        export_table_layout.addWidget(QLabel("Export directory:"))
        export_table_layout.addLayout(export_table_directory_hlayout)
        export_table_layout.addWidget(QLabel('File name:'))
        export_table_layout.addWidget(self.export_filename)
        export_table_layout.addWidget(self.export_table_btn)


        user_input_frame_layout.addWidget(export_figure_grp)
        user_input_frame_layout.addWidget(LineWidget('Horizontal'))
        user_input_frame_layout.addWidget(export_table_grp)
        # user_input_frame_layout.addWidget(info_group)
        # user_input_frame_layout.addWidget(LineWidget('Horizontal'))
        user_input_frame_layout.addStretch()

        # Plot Section
        plot_frame = QFrame()
        plot_frame.setStyleSheet("""
            background-color: #e7e7e7;
            border-top-left-radius: 10px;
            border-top-right-radius: 10px;
        """)


        plot_frame_layout = QVBoxLayout(plot_frame)
        plot_frame_layout.setContentsMargins(0, 0, 0, 0)

        plot_title = QLabel("Figure visualization")
        plot_title.setFixedHeight(30)
        plot_title.setObjectName("TitleBar")
        plot_title.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        plot_title.setContentsMargins(10, 0, 0, 0)
        plot_title.setStyleSheet("""
            background-color: #154E9D;
            color: white;
            font-weight:bold;
            font-size: 16px;
            border-top-left-radius: 10px;
            border-top-right-radius: 10px;
        """)

        plot_frame_layout.addWidget(plot_title)
        plot_frame_layout.addWidget(self.toolbar)
        plot_frame_layout.addWidget(self.canvas)

        top_frame_layout.addWidget(input_section)
        top_frame_layout.addWidget(plot_frame)

        # Table Section
        table_frame = QFrame()
        self.shadow1 = QGraphicsDropShadowEffect(self)
        self.shadow1.setBlurRadius(20)
        self.shadow1.setXOffset(0)
        self.shadow1.setYOffset(0)
        self.shadow1.setColor(QColor(0, 0, 0, 100))
        table_frame.setGraphicsEffect(self.shadow1)

        table_frame.setStyleSheet("QFrame { background-color: transparent; }")

        table_frame_layout = QHBoxLayout(table_frame)
        table_frame_layout.setContentsMargins(20, 20, 20, 20)

        self.styled_table = StyledTable('2D combination table')
        self.styled_table.set_header_label([
            "Set #", "2D Combination", "Predicted 2D peak capacity"
        ])
        self.styled_table.set_default_row_count(10)


        table_frame_layout.addWidget(self.styled_table)

        self.main_splitter = QSplitter(Qt.Vertical, self)
        self.main_splitter.addWidget(top_frame)
        # self.main_splitter.addWidget(table_frame)

        self.main_layout.addWidget(self.main_splitter)

        self.export_figure_directory_btn.clicked.connect(self.create_figure_directory)
        self.export_table_directory_btn.clicked.connect(self.select_export_file_directory)
        self.save_figure_btn.clicked.connect(self.save_figure_list)
        self.export_table_btn.clicked.connect(self.export_tables)

    def init_page(self,om_list):
        self.orthogonality_dict = self.model.get_orthogonality_dict()
        self.plot_utils.set_orthogonality_data(self.model.get_orthogonality_dict())
        data_sets_list = list(self.orthogonality_dict.keys())

        self.figure_list_chklist.add_items(data_sets_list)

        self.figure_type_chklist.clear()
        self.figure_type_chklist.add_items(om_list)

    def create_figure_directory(self):
        directory = QFileDialog.getExistingDirectory(self)
        if directory:
            self.figure_export_directory_lineEdit.setText(directory)

    def select_export_file_directory(self):
        directory = QFileDialog.getExistingDirectory(self)
        if directory:
            self.table_export_directory_lineEdit.setText(directory)

    def export_tables(self):
        select_directory = self.table_export_directory_lineEdit.text()
        file_path = f"{select_directory}/{self.export_filename.text()}"
        table_to_export_list = self.table_selection.get_checked_items()


        with pd.ExcelWriter(file_path, engine="openpyxl") as writer:
            for table_name in table_to_export_list:
                df = self.table_functions_map[table_name]()
                df.to_excel(writer, sheet_name=table_name, index=False)

    def save_figure_list(self):
        chosen_directory = self.figure_export_directory_lineEdit.text()
        chosen_folder_name = f"{chosen_directory}/{self.figure_folder_name_lineEdit.text()}"
        figure_type_list = self.figure_type_chklist.get_checked_item()
        figure_list_chklist = self.figure_list_chklist.get_checked_item()

        if os.path.exists(chosen_directory):
            if not os.path.exists(chosen_folder_name):
                os.mkdir(chosen_folder_name)

            for plot_type in figure_type_list:
                subdirectory_type_name = f"{chosen_folder_name}/{plot_type}"
                if not os.path.exists(subdirectory_type_name):
                    os.mkdir(subdirectory_type_name)
                # self.orthogonality_metric_combo.blockSignals(True)
                # self.orthogonality_metric_combo.setCurrentText(plot_type)
                # self.orthogonality_metric_combo.blockSignals(False)

                for figure_set_nb in figure_list_chklist:
                    self.save_figure(plot_type = plot_type,set_nb=figure_set_nb, dirname=subdirectory_type_name)

    def save_figure(self, plot_type, set_nb, dirname):

        # 1) Grab the Figure from your persistent canvas

        # 2) Totally clear it, then re-add one Axes
        self.plot_utils.clean_figure()
        # ax = fig.add_subplot(111)
        #
        # self.plot_utils.set_axe(ax)

        if self.model.get_status() in ['loaded', 'peak_capacity_loaded']:
            self.plot_utils.plot_scatter(set_number=set_nb, dirname="")

        if plot_type in self.plot_functions_map:
            self.plot_functions_map[plot_type](set_number=set_nb)  # Call the corresponding function

        filename = f"{dirname}/{set_nb}.png"
        self.canvas.figure.savefig(filename, dpi=600, bbox_inches='tight', transparent=True)

    def plot_convex_hull(self,set_number):
        self.plot_utils.plot_convex_hull(set_number=set_number)

    def plot_percent_bin(self,set_number):
        self.plot_utils.plot_percent_bin(set_number=set_number)

    def plot_bin_box(self,set_number):
        self.plot_utils.plot_bin_box(set_number=set_number)

    # Plot methods
    def plot_asterisk(self,set_number):
        self.plot_utils.plot_asterisk(set_number=set_number)

    def plot_linear_reg(self,set_number):
        self.plot_utils.plot_linear_reg(set_number=set_number)

    def plot_percent_fit_xy(self,set_number):
        self.plot_utils.plot_percent_fit_xy(set_number=set_number)

    def plot_percent_fit_yx(self,set_number):
        self.plot_utils.plot_percent_fit_yx(set_number=set_number)

    def plot_conditional_entropy(self,set_number):
        self.plot_utils.plot_conditional_entropy(set_number=set_number)

    def plot_modeling_approach(self,set_number):
        self.plot_utils.plot_modeling_approach(set_number=set_number)



import scipy.cluster.hierarchy as sch

import numpy as np
import pandas as pd
from PySide6.QtWidgets import (
    QVBoxLayout, QHBoxLayout, QDoubleSpinBox,
    QLabel, QCheckBox, QFrame, QGroupBox,
    QSizePolicy, QSplitter, QFormLayout, QScrollArea, QButtonGroup,
    QGraphicsDropShadowEffect
)
from PySide6.QtCore import Qt, QTimer, Signal, QObject, QRunnable, Slot
from PySide6.QtGui import QColor

import seaborn as sns
from matplotlib.patches import Rectangle
from matplotlib.figure import Figure
from matplotlib.backends.backend_qtagg import FigureCanvas

from combo_selector.ui.widgets.custom_toolbar import CustomToolbar
from combo_selector.ui.widgets.line_widget import LineWidget
from combo_selector.ui.widgets.style_table import StyledTable
from combo_selector.ui.widgets.qcombobox_cmap import QComboBoxCmap

METRIC_CORR_MAP = {
    "Convex hull relative area": "Convex hull",
    "Bin box counting": "Bin box",
    "Gilar-Watson method": "Gilar-Watson",
    "Modeling approach": "Mod approach",
    "Conditional entropy": "Cond entropy",  # No short label provided, kept full
    "Pearson Correlation": "Pear corr",
    "Spearman Correlation": "Spea corr",
    "Kendall Correlation": "Kend corr",
    "Asterisk equations": "Asterisk",
    "NND Arithm mean": "NND Amean",
    "NND Geom mean": "NND Gmean",
    "NND Harm mean": "NND Hmean",
    "%FIT": "%FIT",
    "%BIN": "%BIN"
}

class RedundancyCheckPage(QFrame):
    correlation_group_ready = Signal()

    def __init__(self, model=None, title='Unnamed'):
        super().__init__()

        self.heatmap_mask = True
        self.highlight_heatmap_mask = False
        self.model = model
        self.corr_matrix = None
        self.blink_timer = QTimer()
        self.blink_step = 0
        self.blink_ax = None
        self.animations = []
        self.highlighted_ax = None
        self.selected_scatter_collection = None
        self.selected_axe = None
        self.full_scatter_collection = None
        self.selected_set = 'Set 1'
        self.orthogonality_dict = None

        self.model = model
        self.shadow = QGraphicsDropShadowEffect(self)
        self.shadow.setBlurRadius(20)
        self.shadow.setXOffset(0)
        self.shadow.setYOffset(0)
        self.shadow.setColor(QColor(0, 0, 0, 100))

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.main_layout = QVBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)

        self.setFrameShape(QFrame.StyledPanel)
        self.setFrameShadow(QFrame.Raised)

        top_frame = QFrame()
        top_frame.setGraphicsEffect(self.shadow)
        top_frame_layout = QHBoxLayout(top_frame)
        top_frame_layout.setContentsMargins(25, 25, 25, 25)
        top_frame_layout.setSpacing(25)

        user_input_scroll_area = QScrollArea()
        user_input_scroll_area.setFixedWidth(290)
        user_input_scroll_area.setWidgetResizable(True)
        user_input_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        user_input_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        user_input_frame = QFrame()
        user_input_frame.setStyleSheet("background-color: white; border-radius: 10px;")
        user_input_frame.setFixedWidth(280)
        user_input_frame_layout = QVBoxLayout(user_input_frame)
        user_input_frame_layout.setContentsMargins(20, 20, 20, 20)

        user_input_scroll_area.setWidget(user_input_frame)

        input_title = QLabel("Input")
        input_title.setFixedHeight(30)
        input_title.setObjectName("TitleBar")
        input_title.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        input_title.setContentsMargins(10, 0, 0, 0)
        input_title.setStyleSheet("""
            background-color: #154E9D;
            color: white;
            font-weight:bold;
            font-size: 16px;
            border-top-left-radius: 12px;
            border-top-right-radius: 12px;
        """)

        input_section = QFrame()
        input_section.setFixedWidth(290)
        input_layout = QVBoxLayout(input_section)
        input_layout.setSpacing(0)
        input_layout.setContentsMargins(0, 0, 0, 0)
        input_layout.addWidget(input_title)
        input_layout.addWidget(user_input_scroll_area)

        correlation_parameter_group = QGroupBox("Correlation matrix parameter")
        correlation_parameter_group.setStyleSheet("""
            QGroupBox {
                font-size: 14px;
                font-weight: bold;
                background-color: #e7e7e7;
                color: #154E9D;
                border: 1px solid #d0d4da;
                border-radius: 12px;
                margin-top: 25px;
            }
            QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 0px;
                margin-top: -8px;
            }
            QLabel, QCheckBox {
                background-color: transparent;
                color: #3f4c5a;
            }
            QComboBox:hover {
                border: 1px solid #a6b2c0;
            }
            QComboBox::drop-down {
                border: none;
            }
            QComboBox::down-arrow {
                image: url(./drop_down_arrow.png);
            }
        """)

        correlation_parameter_layout = QVBoxLayout()
        form_layout = QFormLayout()

        self.corr_mat_cmap = QComboBoxCmap()
        self.corr_mat_cmap.setCurrentText("BrBG")

        self.correlation_threshold = QDoubleSpinBox()
        self.correlation_threshold.setValue(0.85)

        self.correlation_threshold_tolerance = QDoubleSpinBox()
        self.correlation_threshold_tolerance.setValue(0.0)
        self.correlation_threshold_tolerance.setToolTip("""
            <p>A tolerance of <strong>0</strong> means a metric is only considered correlated if its value is
            <strong>greater than or equal</strong> to the threshold. A tolerance of <strong>0.05</strong> allows metrics
            with correlation values as low as <strong>(threshold - 0.05)</strong> to still be considered correlated.
            For instance, if the threshold is <strong>0.85</strong>, metrics with correlation values down to <strong>0.80</strong>
            will be included.</p>
        """)

        self.highlight_threshold = QCheckBox("Show correlated metric")
        self.highlight_threshold.setChecked(False)

        self.hierarchical_clustering = QCheckBox("Show hierarchical clustering")
        self.hierarchical_clustering.setChecked(False)

        self.lower_triangle_matrix = QCheckBox("Show lower triangle matrix")
        self.lower_triangle_matrix.setChecked(False)

        self.upper_triangle_matrix = QCheckBox("Show upper triangle matrix")
        self.upper_triangle_matrix.setChecked(False)

        self.show_triangle_grp = QButtonGroup()
        self.show_triangle_grp.addButton(self.lower_triangle_matrix)
        self.show_triangle_grp.addButton(self.upper_triangle_matrix)
        self.show_triangle_grp.setExclusive(False)

        form_layout.addRow("Color:", self.corr_mat_cmap)
        form_layout.addRow("Correlation threshold:", self.correlation_threshold)
        form_layout.addRow("Threshold tolerance:", self.correlation_threshold_tolerance)

        correlation_parameter_layout.addLayout(form_layout)
        correlation_parameter_layout.addWidget(self.highlight_threshold)
        correlation_parameter_layout.addWidget(self.hierarchical_clustering)
        correlation_parameter_layout.addWidget(self.lower_triangle_matrix)
        correlation_parameter_layout.addWidget(self.upper_triangle_matrix)

        correlation_parameter_group.setLayout(correlation_parameter_layout)

        # Info Section
        info_page_group = QGroupBox("Info")
        info_page_group.setStyleSheet("""
            QGroupBox {
                font-size: 14px;
                font-weight: bold;
                background-color: #e7e7e7;
                color: #154E9D;
                border: 1px solid #d0d4da;
                border-radius: 12px;
                margin-top: 25px;
            }
            QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 0px;
                margin-top: -8px;
            }
            QLabel {
                background-color: transparent;
                color: #3f4c5a;
            }
        """)

        info_page_layout = QVBoxLayout()
        self.textEdit = QLabel()
        self.textEdit.setTextFormat(Qt.TextFormat.RichText)
        self.textEdit.setWordWrap(True)
        self.textEdit.setText("""
            <p><strong><u>Info 1</u>:</strong><br>
            The table groups metrics that are correlated based on the selected threshold.
            Metrics in the same group have a correlation value <strong>equal to or above</strong> the threshold,
            meaning they behave similarly.</p>
            <p><strong><u>Info 2</u>:</strong><br>
            The <strong>correlation threshold tolerance</strong> allows flexibility in detecting correlated metrics.
            If the absolute difference between a metric’s correlation value and the threshold
            is less than or equal to the tolerance, the metric is considered correlated.</p>
        """)
        info_page_layout.addWidget(self.textEdit)
        info_page_group.setLayout(info_page_layout)

        # Add groups to layout
        user_input_frame_layout.addWidget(correlation_parameter_group)
        user_input_frame_layout.addWidget(LineWidget("Horizontal"))
        user_input_frame_layout.addStretch()
        user_input_frame_layout.addWidget(info_page_group)

        # Plot Frame
        plot_frame = QFrame()
        plot_frame.setStyleSheet("""
            background-color: #e7e7e7;
            border-top-left-radius: 10px;
            border-top-right-radius: 10px;
        """)
        plot_frame_layout = QVBoxLayout(plot_frame)
        plot_frame_layout.setContentsMargins(0, 0, 0, 0)

        plot_title = QLabel("Correlation matrix visualization")
        plot_title.setFixedHeight(30)
        plot_title.setObjectName("TitleBar")
        plot_title.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        plot_title.setContentsMargins(10, 0, 0, 0)
        plot_title.setStyleSheet("""
            background-color: #154E9D;
            color: white;
            font-weight:bold;
            font-size: 16px;
            font-weight: bold;
            padding: 6px 12px;
            border-top-left-radius: 10px;
            border-top-right-radius: 10px;
        """)

        # self.fig = Figure(figsize=(10, 6))
        self.fig = Figure(figsize=(15,15))


        self.canvas = FigureCanvas(self.fig)
        self.toolbar = CustomToolbar(self.canvas)

        # self.fig.set_size_inches(10, 8)  # adjust as needed for your label length
        # self.fig.subplots_adjust(left=0.205,right=.785, bottom=0.365,top=.955)  # increases space for long labels
        self.fig.subplots_adjust(bottom=0.170)  # increases space for long labels

        self._ax = self.canvas.figure.add_subplot(1, 1, 1)
        self._ax.set_box_aspect(1)
        self._ax.set_xlim(0, 1)
        self._ax.set_ylim(0, 1)

        plot_frame_layout.addWidget(plot_title)
        plot_frame_layout.addWidget(self.toolbar)
        plot_frame_layout.addWidget(self.canvas)

        top_frame_layout.addWidget(input_section)
        top_frame_layout.addWidget(plot_frame)

        # Table Frame
        table_frame = QFrame()
        table_frame.setFrameStyle(QFrame.StyledPanel | QFrame.Plain)

        table_frame_layout = QHBoxLayout(table_frame)
        table_frame_layout.setContentsMargins(20, 20, 20, 20)

        self.styled_table = StyledTable("Orthogonality result correlation table")
        self.styled_table.set_header_label(["Group", "Correlated OM"])
        self.styled_table.set_default_row_count(10)

        table_frame_layout.addWidget(self.styled_table)

        self.main_splitter = QSplitter(Qt.Vertical, self)
        self.main_splitter.addWidget(top_frame)
        self.main_splitter.addWidget(table_frame)

        self.main_layout.addWidget(self.main_splitter)

        # Signal Connections
        self.corr_mat_cmap.currentTextChanged.connect(self.update_correlation_matrix_cmap)
        self.correlation_threshold.editingFinished.connect(self.update_correlation_group_table)
        self.correlation_threshold_tolerance.editingFinished.connect(self.update_correlation_group_table)
        self.highlight_threshold.stateChanged.connect(self.highlight_correlation_threshold)
        self.hierarchical_clustering.stateChanged.connect(self.plot_correlation_heat_map)
        self.show_triangle_grp.buttonClicked.connect(self.plot_correlation_heat_map)

    def get_model(self):
        return self.model

    def update_correlation_matrix_cmap(self,cmap):
        quadmesh = self._ax.collections[0]
        quadmesh.set_cmap(cmap)
        self.fig.canvas.draw_idle()

    def init_page(self):
        self.styled_table.clean_table()
        self.styled_table.set_header_label(["Group", "Correlated OM"])

        if self.model.get_orthogonality_metric_corr_matrix_df().empty:
            return

        self.plot_correlation_heat_map()
        self.update_correlation_group_table()

    def plot_correlation_heat_map(self):
        """
        Plots a correlation heatmap of the orthogonality metric correlation matrix.
        Applies optional hierarchical clustering and triangle masking.
        Uses Seaborn for styling and Matplotlib for rendering.
        """
        self.fig.clf()
        self.fig.patch.set_facecolor('white')
        self._ax = self.fig.add_subplot()

        self.corr_matrix = self.model.get_orthogonality_metric_corr_matrix_df().corr()
        cmap = self.corr_mat_cmap.currentText()

        # Map to display names, fall back if missing
        metric_list = [METRIC_CORR_MAP[metric] for metric in list(self.corr_matrix.columns)]

        if self.hierarchical_clustering.checkState() == Qt.Checked:
            self.corr_matrix = self.cluster_corr(self.corr_matrix)

        # Determine triangle mask
        if self.lower_triangle_matrix.checkState() == Qt.Checked:
            self.heatmap_mask = np.triu(np.ones_like(self.corr_matrix, dtype=bool))
        elif self.upper_triangle_matrix.checkState() == Qt.Checked:
            self.heatmap_mask = np.tril(np.ones_like(self.corr_matrix, dtype=bool))
        else:
            self.heatmap_mask = np.zeros_like(self.corr_matrix, dtype=bool)

        # Plot heatmap

        g = sns.heatmap(self.corr_matrix,mask=self.heatmap_mask, vmin=self.corr_matrix.values.min(),
                        vmax=1, square=True, cmap=cmap, linewidths=0.1,
                        annot=True, annot_kws={"fontsize": 6},
                        xticklabels=1, yticklabels=1,ax=self._ax)

        g.set_xticklabels(metric_list, fontsize=7)
        g.set_yticklabels(metric_list, rotation=0, fontsize=7)

        self.highlight_correlation_threshold()

        sns.reset_defaults()
        self.fig.canvas.draw()

    def plot_hierarchical_clustering(self):
        self.fig.clf()
        self._ax = self.canvas.figure.add_subplot()

        self.corr_matrix = self.model.get_orthogonality_metric_corr_matrix_df().corr()

        metric_list = [METRIC_CORR_MAP[metric] for metric in list(self.corr_matrix.columns)]

        if self.hierarchical_clustering.checkState() == Qt.Checked:
            self.corr_matrix = self.cluster_corr(self.corr_matrix)

        cmap = self.corr_mat_cmap.currentText()

        cbar_kws = {"shrink": 1}
        heatmap = sns.heatmap(self.corr_matrix,mask=self.heatmap_mask, vmin=-1, vmax=1, square=True, annot=True, linewidths=0.35,
                              annot_kws={'size':5}, cmap=cmap, ax=self._ax, cbar_kws=cbar_kws)
        sns.set(font_scale=1)
        self._ax.set_xticklabels(metric_list, fontsize=8, rotation=90)
        self._ax.set_yticklabels(metric_list, fontsize=8)

        sns.reset_defaults()

        self.highlight_correlation_threshold()

        self.fig.canvas.draw()
        self.fig.canvas.flush_events()


    def update_correlation_group_table(self):
        threshold = self.correlation_threshold.value()
        tolerance = self.correlation_threshold_tolerance.value()
        self.model.create_correlation_group(threshold=threshold,tol=tolerance)

        correlation_group_table = self.model.get_correlation_group_df()

        self.styled_table.async_set_table_data(correlation_group_table)
        self.styled_table.set_table_proxy()

        self.highlight_correlation_threshold()

        self.correlation_group_ready.emit()


    def highlight_correlation_threshold(self):

        if self.highlight_threshold.checkState() == Qt.Unchecked:
            # Remove all rectangles by iterating over the patches
            for patch in self._ax.patches[:]:  # Iterate over a copy of the list
                patch.remove()

            self.fig.canvas.draw()
            self.fig.canvas.flush_events()
        else:
            threshold = self.correlation_threshold.value()
            tolerance = self.correlation_threshold_tolerance.value()

            # Create a mask: Ignore the diagonal (np.eye) and highlight values above the threshold
            #need to group both conditions with parentheses so Python knows to evaluate them before applying the &:
            self.highlight_heatmap_mask =((self.corr_matrix.abs() >= (threshold - tolerance)) &(~np.eye(len(self.corr_matrix), dtype=bool)))

            self.highlight_heatmap_mask = (~self.heatmap_mask) & self.highlight_heatmap_mask

            # Overlay a border where correlation > threshold (excluding diagonal)
            for i in range(len(self.corr_matrix)):
                for j in range(len(self.corr_matrix)):
                    if self.highlight_heatmap_mask.iloc[i, j]:  # If correlation > threshold and not on the diagonal
                        self._ax.add_patch(Rectangle((j, i), 1, 1, fill=False, edgecolor='red', lw=1))

            self.fig.canvas.draw()
            self.fig.canvas.flush_events()

    def cluster_corr(self,corr_array, inplace=False):
        """
        Rearranges the correlation matrix, corr_array, so that groups of highly
        correlated variables are next to eachother

        Parameters
        ----------
        corr_array : pandas.DataFrame or numpy.ndarray
            a NxN correlation matrix

        Returns
        -------
        pandas.DataFrame or numpy.ndarray
            a NxN correlation matrix with the columns and rows rearranged
        """
        pairwise_distances = sch.distance.pdist(corr_array)
        linkage = sch.linkage(pairwise_distances, method='complete')
        cluster_distance_threshold = pairwise_distances.max() / 2
        idx_to_cluster_array = sch.fcluster(linkage, cluster_distance_threshold,
                                            criterion='distance')
        idx = np.argsort(idx_to_cluster_array)

        if not inplace:
            corr_array = corr_array.copy()

        if isinstance(corr_array, pd.DataFrame):
            return corr_array.iloc[idx, :].T.iloc[idx, :]
        return corr_array[idx, :][:, idx]


        # return result.reset_index(drop=True).set_index(['Variable 1', 'Variable 2'])


import os
import pandas as pd
from PySide6.QtGui import QIcon, QColor, QPixmap
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import (QVBoxLayout, QHBoxLayout, QWidget, QFileDialog, QInputDialog,
    QLabel, QLineEdit, QFrame, QPushButton, QGroupBox, QMessageBox,
    QSizePolicy, QSplitter, QStackedWidget,
    QGraphicsDropShadowEffect, QRadioButton, QButtonGroup
)
from PySide6.QtCore import Qt, QSize, Signal

from combo_selector.ui.widgets.style_table import StyledTable
from combo_selector.ui.widgets.status_icon import Status

PLOT_SIZE = QSize(600, 400)
ICON_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..', '..', 'resources', 'icons')
)

class ImportDataPage(QFrame):

    retention_time_loaded = Signal()
    exp_peak_capacities_loaded = Signal()
    retention_time_normalized = Signal()

    def __init__(self, model=None, title='Unnamed'):
        super().__init__()

        self.model = model

        self.shadow = QGraphicsDropShadowEffect(self)
        self.shadow.setBlurRadius(20)
        self.shadow.setXOffset(0)
        self.shadow.setYOffset(0)
        self.shadow.setColor(QColor(0, 0, 0, 100))

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.main_layout = QVBoxLayout(self)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.setSpacing(0)

        self.setFrameShape(QFrame.StyledPanel)
        self.setFrameShadow(QFrame.Raised)

        top_frame = QFrame()
        top_frame.setFrameShape(QFrame.NoFrame)
        top_frame.setGraphicsEffect(self.shadow)
        top_frame.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        top_frame_layout = QHBoxLayout(top_frame)
        top_frame_layout.setContentsMargins(25, 25, 25, 25)
        top_frame_layout.setSpacing(25)

        # ====== INFORMATION SECTION ======
        separation_space_scaling_title = QLabel("Separation Space Scaling")
        separation_space_scaling_title.setFixedHeight(30)
        separation_space_scaling_title.setObjectName("TitleBar")
        separation_space_scaling_title.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        separation_space_scaling_title.setContentsMargins(10, 0, 0, 0)
        separation_space_scaling_title.setStyleSheet("""
            background-color: #154E9D;
            color: white;
            font-weight:bold;
            font-size: 15px;
            border-top-left-radius: 12px;
            border-top-right-radius: 12px;
            margin-bottom: 0px;
        """)


        user_input_frame = QFrame()
        user_input_frame.setStyleSheet("background-color: white; border-radius: 0px;")
        user_input_frame_layout = QHBoxLayout(user_input_frame)
        user_input_frame_layout.setContentsMargins(40, 40, 40, 40)

        # --- GROUP BOX ---
        scaling_method_group = QGroupBox("Select scaling method")
        scaling_method_group.setStyleSheet("""
            QGroupBox {
                font-size: 14px;
                font-weight: bold;
                background-color: white;
                color: #154E9D;
                border: 1px solid #d0d4da;
                border-radius: 12px;
                margin-top: 25px;
                
            }

          QGroupBox::title {
                subcontrol-origin: margin;
                subcontrol-position: top left;
                padding: 0px;
                margin-top: -8px;
            }

            QLabel {
                background-color: transparent;
                color: #2C3E50;
                font-family: "Segoe UI";
                font-size: 13px;
            }

           QRadioButton, QCheckBox {
                background-color: transparent;
            color: #2C3E50;              /* dark slate navy */
            
            }

            QPushButton {
                background-color: #d5dcf9;
                color: #2C3346;
                border: none;
                border-radius: 6px;
                padding: 10px;
                font-size: 14px;
                font-weight: 600;
            }

            QPushButton:hover {
                background-color: #bcc8f5;
            }

            QPushButton:pressed {
                background-color: #8fa3ef;
            }

            QPushButton:disabled {
                background-color: #E5E9F5;
                color: #FFFFFF;
            }
        """)

        # --- MAIN LAYOUT ---
        scaling_method_layout = QVBoxLayout()

        # --- Normalize Button ---
        self.normalize_btn = QPushButton("Normalize data")

        # --- RADIO BUTTONS ---
        self.min_max_scaling_btn = QRadioButton("Min-Max scaling")
        self.min_max_scaling_btn.setObjectName('min_max')
        self.min_max_scaling_btn.setChecked(True)

        self.void_max_scaling_btn = QRadioButton("Void – Max scaling")
        self.void_max_scaling_btn.setObjectName('void_max')

        self.wosel_btn = QRadioButton("WOSEL")
        self.wosel_btn.setObjectName('wosel')

        # --- GROUP RADIO BUTTONS ---
        self.radio_button_group = QButtonGroup()
        self.radio_button_group.addButton(self.min_max_scaling_btn)
        self.radio_button_group.addButton(self.void_max_scaling_btn)
        self.radio_button_group.addButton(self.wosel_btn)
        self.radio_button_group.setExclusive(True)

        # --- RADIO LAYOUT ---
        radio_layout = QVBoxLayout()
        radio_layout.addWidget(self.min_max_scaling_btn)
        radio_layout.addWidget(self.void_max_scaling_btn)
        radio_layout.addWidget(self.wosel_btn)
        # radio_layout.addStretch()

        radio_widget = QWidget()
        radio_widget.setStyleSheet("background-color: transparent;")
        radio_widget.setLayout(radio_layout)
        radio_widget.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        radio_widget.setMaximumWidth(200)

        # --- SVG STACK ---
        self.scaling_method_svg_qstack = QStackedWidget()
        # self.scaling_method_svg_qstack.setFixedHeight(100)

        self.norm_min_max_svg = QSvgWidget()

        self.norm_min_max_svg.load(os.path.join(ICON_PATH, 'norm_min_max.svg'))
        self.norm_min_max_svg.setAttribute(Qt.WA_TranslucentBackground)
        self.norm_min_max_svg.setStyleSheet("background: transparent;")
        self.norm_min_max_svg.renderer().setAspectRatioMode(Qt.KeepAspectRatio)

        self.norm_void_max_svg = QSvgWidget()
        self.norm_void_max_svg.setAttribute(Qt.WA_TranslucentBackground)
        self.norm_void_max_svg.setStyleSheet("background: transparent;")
        self.norm_void_max_svg.load(os.path.join(ICON_PATH, 'norm_void_max.svg'))
        self.norm_void_max_svg.renderer().setAspectRatioMode(Qt.KeepAspectRatio)

        self.norm_wosel_svg = QSvgWidget()
        self.norm_wosel_svg.setAttribute(Qt.WA_TranslucentBackground)
        self.norm_wosel_svg.setStyleSheet("background: transparent;")
        self.norm_wosel_svg.load(os.path.join(ICON_PATH, 'norm_wosel.svg'))
        self.norm_wosel_svg.renderer().setAspectRatioMode(Qt.KeepAspectRatio)

        self.scaling_method_svg_qstack.addWidget(self.norm_min_max_svg)
        self.scaling_method_svg_qstack.addWidget(self.norm_void_max_svg)
        self.scaling_method_svg_qstack.addWidget(self.norm_wosel_svg)

        # --- SVG CONTAINER ---
        svg_container = QWidget()
        # svg_container.setFixedSize(QSize(218*1.2, 93*1.2))
        svg_layout = QVBoxLayout()
        svg_layout.addStretch()  # Top stretch
        svg_layout.addWidget(self.scaling_method_svg_qstack, alignment=Qt.AlignCenter)  # Center SVG horizontally
        svg_layout.addStretch()  # Bottom stretch
        svg_layout.setContentsMargins(0, 0, 0, 0)
        svg_container.setLayout(svg_layout)
        svg_container.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)  # Allow vertical growth

        # --- HORIZONTAL LAYOUT (Left + Right) ---
        hlayout = QHBoxLayout()
        hlayout.setAlignment(Qt.AlignVCenter)  # Ensure row is vertically centered in parent
        hlayout.addWidget(radio_widget)
        hlayout.addWidget(svg_container)

        # --- VERTICAL CENTERING ---
        centered_layout = QVBoxLayout()
        centered_layout.addStretch()
        centered_layout.addLayout(hlayout)
        centered_layout.addStretch()

        # --- FINAL ASSEMBLY ---
        scaling_method_layout.addLayout(centered_layout)
        scaling_method_layout.addSpacing(10)
        scaling_method_layout.addWidget(self.normalize_btn, alignment=Qt.AlignCenter)

        scaling_method_group.setLayout(scaling_method_layout)

        user_input_frame_layout.addWidget(scaling_method_group)

        normalization_section = QFrame()
        normalization_layout = QVBoxLayout(normalization_section)
        normalization_layout.setSpacing(0)
        normalization_layout.setContentsMargins(0, 0, 0, 0)
        normalization_layout.addWidget(separation_space_scaling_title)
        normalization_layout.addWidget(user_input_frame)

        # ====== DATA FILES SECTION (clean and squared) ======
        import_data_inner_frame = QFrame()

        import_data_inner_frame.setStyleSheet("""
            QFrame {
                background-color: #ffffff;
                border: none;
            }

            QLabel {
                color: #3f4c5a;
                font-size: 13px;
            }

            QLineEdit {
                background-color: #f5f6f7;
                border: 1px solid #d1d6dd;
                border-radius: 4px;
                padding: 4px 6px;
                font-size: 12px;
            }

            QPushButton {
                background-color: #f1f3f6;
                border: none;
                border-radius: 3px;
                padding: 4px 10px;
                font-size: 12px;
            }

            QPushButton:hover {
                background-color: #e1e6ec;
            }
        """)

        import_data_inner_layout = QVBoxLayout(import_data_inner_frame)
        # import_data_inner_layout.setSpacing(10)
        # import_data_inner_layout.setContentsMargins(20, 20, 20, 20)

        # Retention time
        self.add_ret_time_btn = QPushButton(QIcon("Circle-icons-folder.png"), 'Import')
        self.add_ret_time_btn.setFixedHeight(24)
        self.add_ret_time_filename = QLineEdit()
        self.add_ret_time_filename.setFixedHeight(24)
        self.ret_time_import_status = Status()
        norm_layout = QHBoxLayout()
        norm_layout.addWidget(self.add_ret_time_filename)
        norm_layout.addWidget(self.add_ret_time_btn)
        norm_layout.addWidget(self.ret_time_import_status)

        # 1D peak capacity
        self.add_2D_peak_data_btn = QPushButton(QIcon("Circle-icons-folder.png"), 'Import')
        self.add_2D_peak_data_btn.setFixedHeight(24)
        self.add_2D_peak_data_linedit = QLineEdit()
        self.add_2D_peak_data_linedit.setFixedHeight(24)
        self.twoD_peak_status = Status()
        peak_layout = QHBoxLayout()
        peak_layout.addWidget(self.add_2D_peak_data_linedit)
        peak_layout.addWidget(self.add_2D_peak_data_btn)
        peak_layout.addWidget(self.twoD_peak_status)



        # Retention time
        self.add_void_time_btn = QPushButton(QIcon("Circle-icons-folder.png"), 'Import')
        self.add_void_time_btn.setFixedHeight(24)
        self.add_void_time_filename = QLineEdit()
        self.add_void_time_filename.setFixedHeight(24)
        self.void_time_import_status = Status()
        self.void_time_widget = QWidget()
        self.void_time_widget.setVisible(False)
        void_time_layout = QHBoxLayout(self.void_time_widget)
        void_time_layout.setContentsMargins(0, 0, 0, 0)
        void_time_layout.addWidget(self.add_void_time_filename)
        void_time_layout.addWidget(self.add_void_time_btn)
        void_time_layout.addWidget(self.void_time_import_status)
        self.void_time_label = QLabel("Void time:")
        self.void_time_label.setVisible(False)


        # Retention time
        self.add_gradient_end_time_btn = QPushButton(QIcon("Circle-icons-folder.png"), 'Import')

        self.add_gradient_end_time_btn.setFixedHeight(24)
        self.add_gradient_end_time_filename = QLineEdit()
        self.add_gradient_end_time_filename.setFixedHeight(24)
        self.gradient_end_time_import_status = Status()
        self.gradient_end_time_widget = QWidget()
        self.gradient_end_time_widget.setVisible(False)
        gradient_end_time_layout = QHBoxLayout(self.gradient_end_time_widget)
        gradient_end_time_layout.setContentsMargins(0, 0, 0, 0)
        gradient_end_time_layout.addWidget(self.add_gradient_end_time_filename)
        gradient_end_time_layout.addWidget(self.add_gradient_end_time_btn)
        gradient_end_time_layout.addWidget(self.gradient_end_time_import_status)
        self.gradient_end_time_label = QLabel("Gradient end time:")
        self.gradient_end_time_label.setVisible(False)

        # Add widgets to layout
        import_data_inner_layout.addStretch()
        import_data_inner_layout.addWidget(QLabel("Retention times:"))
        import_data_inner_layout.addLayout(norm_layout)
        import_data_inner_layout.addWidget(QLabel("Experimental 1D peak capacities:"))
        import_data_inner_layout.addLayout(peak_layout)
        import_data_inner_layout.addWidget(self.void_time_label)
        import_data_inner_layout.addWidget(self.void_time_widget)
        import_data_inner_layout.addWidget(self.gradient_end_time_label)
        import_data_inner_layout.addWidget(self.gradient_end_time_widget)
        import_data_inner_layout.addStretch()

        # === Title Bar + Wrapping Frame ===
        data_file_title = QLabel("Data import")
        data_file_title.setFixedHeight(30)
        data_file_title.setObjectName("TitleBar")
        data_file_title.setAlignment(Qt.AlignVCenter | Qt.AlignLeft)
        data_file_title.setContentsMargins(10, 0, 0, 0)
        data_file_title.setStyleSheet("""
            background-color: #154E9D;
            color: white;
            font-weight:bold;
            font-size: 15px;
            border-top-left-radius: 12px;
            border-top-right-radius: 12px;
            margin-bottom: 0px;
        """)

        import_data_frame = QFrame()
        import_data_layout = QVBoxLayout(import_data_frame)
        import_data_layout.setSpacing(0)
        import_data_layout.setContentsMargins(0, 0, 0, 0)
        import_data_layout.addWidget(data_file_title)
        import_data_layout.addWidget(import_data_inner_frame)

        # ====== NORMALIZED TABLE SECTION (optional below) ======
        table_frame = QFrame()
        self.shadow1 = QGraphicsDropShadowEffect(self)
        self.shadow1.setBlurRadius(20)
        self.shadow1.setXOffset(0)
        self.shadow1.setYOffset(0)
        self.shadow1.setColor(QColor(0, 0, 0, 100))
        table_frame.setGraphicsEffect(self.shadow1)
        table_frame.setStyleSheet("QFrame { background-color: transparent; }")
        table_frame_layout = QHBoxLayout(table_frame)
        table_frame_layout.setContentsMargins(20, 20, 20, 20)

        self.normalized_data_table = StyledTable("Normalized Retention time table")
        self.normalized_data_table.set_header_label(["Peak #", "Condition 1", "Condition 2", "...", "Condition n"])
        self.normalized_data_table.set_default_row_count(10)
        table_frame_layout.addWidget(self.normalized_data_table)

        # ====== FINAL ASSEMBLY ======
        top_frame_layout.addWidget(import_data_frame,50)
        top_frame_layout.addWidget(normalization_section,50)

        self.main_splitter = QSplitter(Qt.Orientation.Vertical, self)
        self.main_splitter.addWidget(top_frame)
        self.main_splitter.addWidget(table_frame)

        self.main_layout.addWidget(self.main_splitter)

        self.radio_button_group.buttonClicked.connect(self.change_norm_svg)
        self.normalize_btn.clicked.connect(self.normalize_retention_time)
        self.add_ret_time_btn.clicked.connect(self.load_retention_data)
        self.add_2D_peak_data_btn.clicked.connect(self.load_experimental_peak_capacities)
        self.add_void_time_btn.clicked.connect(self.load_void_time_data)
        self.add_gradient_end_time_btn.clicked.connect(self.load_gradient_end_time_data)

    # ==============================
    # Data Handling & Selection
    # =============================*
    #

    def change_norm_svg(self):
        button_checked = self.radio_button_group.checkedButton()
        method = button_checked.objectName()

        if method == 'min_max':
            self.scaling_method_svg_qstack.setCurrentIndex(0)
            self.void_time_widget.setVisible(False)
            self.gradient_end_time_widget.setVisible(False)
            self.void_time_label.setVisible(False)
            self.gradient_end_time_label.setVisible(False)

        if method == 'void_max':
            self.scaling_method_svg_qstack.setCurrentIndex(1)
            self.void_time_widget.setVisible(True)
            self.void_time_label.setVisible(True)
            self.gradient_end_time_widget.setVisible(False)
            self.gradient_end_time_label.setVisible(False)

        if method == 'wosel':
            self.scaling_method_svg_qstack.setCurrentIndex(2)
            self.void_time_widget.setVisible(True)
            self.void_time_label.setVisible(True)
            self.gradient_end_time_widget.setVisible(True)
            self.gradient_end_time_label.setVisible(True)


    def normalize_retention_time(self):
        button_checked = self.radio_button_group.checkedButton()
        method = button_checked.objectName()
        self.model.normalize_retention_time(method)

        data = self.model.get_normalized_retention_time_df()


        self.normalized_data_table.async_set_table_data(data)

        self.retention_time_normalized.emit()


    def load_retention_data(self) -> None:
        """
        Opens a file dialog to select and load an Excel file containing normalized retention data.

        - Allows the user to select an Excel file and choose a sheet.
        - Loads the selected sheet into the `Orthogonality` model.
        - Updates the UI components accordingly.
        - Displays an error message and updates the GUI if loading fails.

        Raises:
            - FileNotFoundError: If the user cancels the file selection.
            - ValueError: If the selected sheet is invalid.
            - Exception: For unexpected errors.
        """
        file_path, _ = QFileDialog.getOpenFileName(self, "Open Excel File", "", "Excel Files (*.xlsx *.xls)")

        if not file_path:
            return  # User canceled the selection

        try:
            sheet_names = pd.ExcelFile(file_path, engine="openpyxl").sheet_names
            selected_sheet, ok = QInputDialog.getItem(self, "Select Sheet", "Choose a sheet:", sheet_names,
                                                      editable=False)

            if not ok:
                raise ValueError("No sheet selected")

            self.model.load_2d_set(filepath=file_path, sheetname=selected_sheet)

            if self.model.get_status() == "error":
                self.ret_time_import_status.set_error()
                QMessageBox.critical(self, "Error", "Failed to load the data. Please check the file format.")
                return

            # Successful load: update UI
            self.ret_time_import_status.set_valid()
            self.add_ret_time_filename.setText(file_path)

            data = self.model.get_retention_time_df()

            self.normalized_data_table.set_header_label(list(data.columns))

            self.normalized_data_table.async_set_table_data(data)

            self.retention_time_loaded.emit()

        except ValueError as e:
            self.ret_time_import_status.set_error()  # Ensure GUI shows failure status
            QMessageBox.warning(self, "Warning", str(e))
        except Exception as e:
            self.ret_time_import_status.set_error()  # Ensure GUI shows failure status
            QMessageBox.critical(self, "Error", f"An unexpected error occurred:\n{str(e)}")

    def load_experimental_peak_capacities(self):
        fileName = QFileDialog.getOpenFileName(self, "Open Excel File", "", "Excel Files (*.xlsx *.xls)")
        if fileName[0]:
            try:
                sheet_names_list = pd.ExcelFile(fileName[0], engine='openpyxl').sheet_names
                sheet, ok = QInputDialog.getItem(self, 'Select excel sheet', 'select sheet', sheet_names_list)
            except:
                ok = False

            if ok:
                self.model.load_data_frame_2d_peak(filepath=fileName[0],
                                                   sheetname=sheet)

                status = self.model.get_status()

                if status == 'error':
                    self.twoD_peak_status.set_error()
                else:
                    self.twoD_peak_status.set_valid()
                    self.add_2D_peak_data_linedit.setText(fileName[0])
                    self.exp_peak_capacities_loaded.emit()
            else:
                self.twoD_peak_status.set_error()

    def load_gradient_end_time_data(self):
        fileName = QFileDialog.getOpenFileName(self, "Open Excel File", "", "Excel Files (*.xlsx *.xls)")
        if fileName[0]:
            try:
                sheet_names_list = pd.ExcelFile(fileName[0], engine='openpyxl').sheet_names
                sheet, ok = QInputDialog.getItem(self, 'Select excel sheet', 'select sheet', sheet_names_list)
            except:
                ok = False

            if ok:
                self.model.load_gradient_end_time(filepath=fileName[0],
                                                   sheetname=sheet)

                status = self.model.get_status()

                if status == 'error':
                    self.gradient_end_time_import_status.set_error()
                else:
                    self.gradient_end_time_import_status.set_valid()
                    self.add_gradient_end_time_filename.setText(fileName[0])
            else:
                self.gradient_end_time_import_status.set_error()

    def load_void_time_data(self):
        fileName = QFileDialog.getOpenFileName(self, "Open Excel File", "", "Excel Files (*.xlsx *.xls)")
        if fileName[0]:
            try:
                sheet_names_list = pd.ExcelFile(fileName[0], engine='openpyxl').sheet_names
                sheet, ok = QInputDialog.getItem(self, 'Select excel sheet', 'select sheet', sheet_names_list)
            except:
                ok = False

            if ok:
                self.model.load_void_time(filepath=fileName[0],
                                                   sheetname=sheet)

                status = self.model.get_status()

                if status == 'error':
                    self.void_time_import_status.set_error()
                else:
                    self.void_time_import_status.set_valid()
                    self.add_void_time_filename.setText(fileName[0])
            else:
                self.void_time_import_status.set_error()






from PySide6.QtWidgets import QTreeWidget, QTreeWidgetItem, QVBoxLayout, QWidget
from PySide6.QtCore import Qt


class CheckableTreeList(QWidget):
    def __init__(self,item_list=None):
        super().__init__()

        self.tree = QTreeWidget()
        self.tree.setHeaderHidden(True)

        # Create "Select all" parent item
        self.__init_tree()

        self.add_items(item_list)

        self.tree.itemChanged.connect(self.handle_item_changed)

        layout = QVBoxLayout()
        layout.addWidget(self.tree)
        self.setLayout(layout)

    def clear(self):
        self.tree.clear()
        self.__init_tree()

    def __init_tree(self):
        # List of child items
        self.children = []
        # Create "Select all" parent item
        self.parent_item = QTreeWidgetItem(self.tree, ["Select all"])
        self.parent_item.setFlags(self.parent_item.flags() | Qt.ItemIsUserCheckable)
        self.parent_item.setCheckState(0, Qt.Unchecked)
        self.parent_item.setExpanded(True)  # Set expanded by default

    def add_items(self,item_list):
        if item_list:
            for item in item_list:
                child = QTreeWidgetItem(self.parent_item, [item])
                child.setFlags(child.flags() | Qt.ItemIsUserCheckable)
                child.setCheckState(0, Qt.Unchecked)
                self.children.append(child)

        self.parent_item.setExpanded(True)


    def handle_item_changed(self, item, column):
        if item is self.parent_item and item.checkState(0) in (Qt.Checked, Qt.Unchecked):
            state = item.checkState(0)
            self.tree.blockSignals(True)  # Prevent recursive triggers
            for child in self.children:
                child.setCheckState(0, state)
            self.tree.blockSignals(False)
        elif item in self.children:
            checked_count = sum(child.checkState(0) == Qt.Checked for child in self.children)
            unchecked_count = sum(child.checkState(0) == Qt.Unchecked for child in self.children)
            partially_checked = any(child.checkState(0) == Qt.PartiallyChecked for child in self.children)
            self.tree.blockSignals(True)
            if checked_count == len(self.children):
                self.parent_item.setCheckState(0, Qt.Checked)
            elif unchecked_count == len(self.children):
                self.parent_item.setCheckState(0, Qt.Unchecked)
            else:
                self.parent_item.setCheckState(0, Qt.PartiallyChecked)
            self.tree.blockSignals(False)

    def get_checked_items(self):
        return [child.text(0) for child in self.children if child.checkState(0) == Qt.Checked]

    def get_items(self):
        return [child.text(0) for child in self.children]


    def unchecked_all(self):
        self.tree.blockSignals(True)
        self.parent_item.setCheckState(0, Qt.Unchecked)

        for child in self.children:
            child.setCheckState(0, Qt.Unchecked)

        self.tree.blockSignals(False)


#############################################################################################
# CREATOR:  ANJAL.P
#           J#
# ON:       2020 SEP.                                                                       #
# AIM:      To Extend the capability of the PySide2 and PyQt5 Python library with easy to   #
#           use extension containing commonly used widgets which is not natively supported  #
#           by the Qt Frame work (or atleast for Python version of Qt).                     #
# VERSION:  v2.0.0 (update version of v1.0.0 with migration from PySide2 to Pyside6)                                               #
# NOTES:    CLASS : RoundProgressBar : Can be accessed by : importing                       #
#           from PySide2extn.RoundProgressBar import RoundProgressBar                       #
# REFER:    Github: https://github.com/anjalp/PySide2extn                                   #
#############################################################################################
import sys

from PySide6 import QtWidgets, QtCore
from PySide6.QtCore import Qt, QSize
from PySide6.QtGui import QColor, QPainter, QPen, QPaintEvent, QFont
from PySide6.QtCore import QTimer



class RoundProgressBar(QtWidgets.QWidget):

    def __init__(self, parent=None):
        super(RoundProgressBar, self).__init__(parent)

        self.setStyleSheet("background: transparent;")
        self.setFixedSize(100,100)
        # Inside __init__ of RoundProgressBar
        self.hide_timer = QTimer()
        self.hide_timer.setSingleShot(True)
        self.hide_timer.timeout.connect(self.hide)
        self.hide()  # Initially hidden

        self.positionX = 0
        self.positionY = 0
        self.posFactor = 0
        self.textOffset = 20

        self.rpb_minimumSize = (0, 0)
        self.rpb_maximumSize = (0, 0)
        self.rpb_dynamicMin = True
        self.rpb_dynamicMax = True
        self.rpb_Size = 0
        self.sizeFactor = 0

        self.rpb_maximum = 100
        self.rpb_minimum = 0

        self.rpb_type = self.barStyleFlags.Donet
        self.startPosition = self.startPosFlags.North
        self.rpb_direction = self.rotationFlags.Clockwise

        self.rpb_textType = self.textFlags.Percentage
        self.rpb_textColor = '#154E9D'
        self.rpb_textWidth = self.rpb_Size / 8
        self.rpb_textFont = 'Segoe UI'
        self.rpb_textValue = '0%'
        self.rpb_textStatus = 'Status'
        self.rpb_textRatio = 8
        self.textFactorX = 0
        self.textFactorY = 0
        self.dynamicText = True
        self.rpb_textActive = True

        self.lineWidth = 10
        self.pathWidth = 10
        self.rpb_lineStyle = self.lineStyleFlags.SolidLine
        self.rpb_lineCap = self.lineCapFlags.SquareCap
        self.lineColor = '#154E9D'
        self.pathColor = '#e7e7e7'

        self.rpb_circleColor = '#e7e7e7'
        self.rpb_circleRatio = 0.8
        self.rpb_circlePosX = 0
        self.rpb_circlePosY = 0

        self.rpb_pieColor = (200, 200, 200)
        self.rpb_pieRatio = 1
        self.rpb_piePosX = 0
        self.rpb_piePosY = 0

        self.rpb_value = 0

        if self.rpb_dynamicMin:
            self.setMinimumSize(QSize(self.lineWidth * 6 + self.pathWidth * 6, self.lineWidth * 6 + self.pathWidth * 6))

    # ------------------------------------------------------CLASS ENUMERATORS
    class lineStyleFlags:
        SolidLine = Qt.SolidLine
        DotLine = Qt.DotLine
        DashLine = Qt.DashLine

    class lineCapFlags:
        SquareCap = Qt.SquareCap
        RoundCap = Qt.RoundCap

    class barStyleFlags:
        Donet = 0
        Line = 1
        Pie = 2
        Pizza = 3
        Hybrid1 = 4
        Hybrid2 = 5

    class rotationFlags:
        Clockwise = -1
        AntiClockwise = 1

    class textFlags:
        Value = 0
        Percentage = 1

    class startPosFlags:
        North = 90 * 16
        South = -90 * 16
        East = 0 * 16
        West = 180 * 16

    # ------------------------------------------------------METHODS FOR CHANGING THE PROPERTY OF THE RoundProgressBar :SOLTS

    def rpb_set_minimum_size(self, width, height):
        """
        Minimum Size of the Widget
        ...

        Parameters
        --------------

        width : int
            width of the Widget

        height : int
            height of the Widget

        Raises
        --------------
        Exception : Sorry Width/Height should be an int
        """

        if type(width) != type(5) or type(height) != type(5):
            raise Exception('Sorry Width/Height should be an int')
            return
        self.rpb_dynamicMin = False
        self.setMinimumSize(width, height)
        self.rpb_minimumSize = (width, height)
        self.update()

    def rpb_setMaximumSize(self, width, height):
        """
        Maximum Size of the Widget
        ...

        Parameters
        --------------

        width : int
            width of the Widget

        height : int
            height of the Widget

        Raises
        --------------
        Exception : Sorry Width/Height should be an int
        """

        if type(width) != type(5) or type(height) != type(5):
            raise Exception('Sorry Width/Height should be an int')
            return
        self.rpb_dynamicMax = False
        self.setMaximumSize(width, height)
        self.rpb_maximumSize = (width, height)
        self.update()

    def rpb_setMaximum(self, maximum):
        """
        Maximum Value of the Progressbar
        ...

        Parameters
        --------------

        maximum : int
            Maximum value of the round progress bar

        Raises
        --------------
        Exception : Maximum and Minimum cannot be the Same
        """

        if self.rpb_minimum == maximum:  # FOR AVOIDING DIVISION BY ZERO ERROR IN FUTURE
            raise Exception("Maximum and Minimum cannot be the Same")
            return
        if self.rpb_maximum != maximum:
            self.rpb_maximum = maximum
            self.update()

    def rpb_setMinimum(self, minimum):
        """
        Minimum Value of the Progressbar
        ...

        Parameters
        --------------

        minimum : int
            Minimum value of the round progress bar

        Raises
        --------------
        Exception : Maximum and Minimum cannot be the Same
        """

        if self.rpb_minimum == maximum:  # FOR AVOIDING DIVISION BY ZERO ERROR IN FUTURE
            raise Exception("Maximum and Minimum cannot be the Same")
            return
        if self.rpb_minimum != minimum:
            self.rpb_minimum = minimum
            self.update()

    def rpb_setRange(self, maximum, minimum):
        """
        Range include the maximum and the minimum in one go.
        ...

        Parameters
        --------------

        maximum : int
            Maximum value of the round progress bar

        minimum : int
            Minimum value for the round progress bar

        Raises
        --------------
        none
        """

        if minimum > maximum:
            maximum, minimum = minimum, maximum
        if self.rpb_maximum != maximum:
            self.rpb_maximum = maximum
        if self.rpb_minimum != minimum:
            self.rpb_minimum = minimum
        self.update()

    def rpb_setInitialPos(self, pos):
        """
        Starting position of the round progress bar
        ...

        Parameters
        --------------

        pos : String
            Position string: 'North', 'South', 'East' and 'West'

        Raises
        --------------
        ValueError : Maximum and Minimum cannot be the Same
        """

        if pos == 'North':
            self.startPosition = self.startPosFlags.North
        elif pos == 'South':
            self.startPosition = self.startPosFlags.South
        elif pos == 'East':
            self.startPosition = self.startPosFlags.East
        elif pos == 'West':
            self.startPosition = self.startPosFlags.West
        else:
            raise Exception("Initial Position String can be: 'South', 'North'")
            return

    def rpb_setValue(self, value):
        """
        Set progress value and manage visibility.
        """

        # Clamp value
        if value > self.rpb_maximum:
            value = self.rpb_maximum
        elif value < self.rpb_minimum:
            value = self.rpb_minimum

        # Avoid unnecessary updates
        if self.rpb_value == value:
            return

        # Cancel any pending hide if value is < 100
        if value < self.rpb_maximum and self.hide_timer.isActive():
            self.hide_timer.stop()

        # Convert and update
        self.convertInputValue(value)
        self.update()

        # Manage visibility
        if value == self.rpb_minimum:
            self.hide()
        elif value < self.rpb_maximum:
            self.show()
        elif value == self.rpb_maximum:
            # Show in case it was hidden due to flicker
            self.show()
            self.hide_timer.start(800)  # Delay before hiding after reaching 100%

    def rpb_reset(self):
        """
        Reset the progress bar to 0%
        ...

        Parameters
        --------------
        none

        Raises
        --------------
        none
        """

        RoundProgressBar.convertInputValue(self, self.rpb_minimum)
        self.update()

    def rpb_setGeometry(self, posX, posY):
        """
        Set the X and Y position of the round progress bar.
        ...

        Parameters
        --------------

        posX : int
            The position of the round progress bar in int for X axis.

        posY : int
            The position of the round progress bar in int for Y axis.

        Raises
        --------------
        none
        """

        if self.positionX != posX:
            self.positionX = posX
        if self.positionY != posY:
            self.positionY = posY
        self.update()

    def rpb_setLineWidth(self, width):
        """
        Line Width of the line in round progress bar.
        ...

        Parameters
        --------------

        width: int
            Line width corresponding to the width in px.

        Raises
        --------------
        Exception: Line Width should be in int
        """

        if type(width) != type(5):
            raise Exception('Line Width should be in int')
            return
        if self.lineWidth != width:
            self.lineWidth = width
            self.update()

    def rpb_setLineColor(self, rgb):
        """
        Line Color of the progress bar.
        ...

        Parameters
        --------------

        rgb: tuple: (R, G, B)
            Color is passed as a tuple of values for red, blue and green in the order: (R, G, B)

        Raises
        --------------
        Exception: Line Color accepts a tuple: (R, G, B).
        """

        if type(rgb) != type(()):
            raise Exception("Line Color accepts a tuple: (R, G, B).")
            return
        if self.lineColor != rgb:
            self.lineColor = rgb
            self.update()

    def rpb_setPathColor(self, rgb):
        """
        Path Color settings.
        ...

        Parameters
        --------------

        rgb: tuple: (R, G, B)
            Color is passed as a tuple of values for red, blue and green in the order: (R, G, B)

        Raises
        --------------
        Exception: Path Color accepts a tuple: (R, G, B).
        """

        if type(rgb) != type(()):
            raise Exception("Path Color accepts a tuple: (R, G, B).")
            return
        if self.pathColor != rgb:
            self.pathColor = rgb
            self.update()

    def rpb_setPathWidth(self, width):
        """
        Path width settings.
        ...

        Parameters
        --------------

        width: int
            Width of the path in px

        Raises
        --------------
        Exception: Line Width should be in int
        """

        if type(width) != type(5):
            raise Exception('Path Width should be in int')
            return
        if self.pathWidth != width:
            self.pathWidth = width
            self.update()

    def rpb_setDirection(self, direction):
        """
        Direction of rotation of the progress bar.
        ...

        Parameters
        --------------

        direction: string
            string can be: 'AntiClockwise' or 'Clockwise'. Default: 'Clockwise'.

        Raises
        --------------
        Exception: Direction can only be: 'Clockwise' and 'AntiClockwise'
        """

        if direction == 'Clockwise' or direction == -1:
            self.rpb_direction = self.rotationFlags.Clockwise
        elif direction == 'AntiClockwise' or direction == 1:
            self.rpb_direction = self.rotationFlags.AntiClockwise
        else:
            raise Exception("Direction can only be: 'Clockwise' and 'AntiClockwise' and Not: " + str(direction))
            return
        self.update()

    def rpb_setBarStyle(self, style):
        """
        Bar Style of the progress bar.
        ...

        Parameters
        --------------

        style: String
            String of the styles of the progress bar: 'Donet', 'Pie', 'line', 'Hybrid1', 'Hybrid2', 'Pizza'

        Raises
        --------------
        Exception: Round Progress Bar has only the following styles: 'Line', 'Donet', 'Hybrid1', 'Pizza', 'Pie' and 'Hybrid2'
        """

        if style == 'Donet':
            self.rpb_type = self.barStyleFlags.Donet
        elif style == 'Line':
            self.rpb_type = self.barStyleFlags.Line
        elif style == 'Pie':
            self.rpb_type = self.barStyleFlags.Pie
        elif style == 'Pizza':
            self.rpb_type = self.barStyleFlags.Pizza
        elif style == 'Hybrid1':
            self.rpb_type = self.barStyleFlags.Hybrid1
        elif style == 'Hybrid2':
            self.rpb_type = self.barStyleFlags.Hybrid2
        else:
            raise Exception(
                "Round Progress Bar has only the following styles: 'Line', 'Donet', 'Hybrid1', 'Pizza', 'Pie' and 'Hybrid2'")
            return
        self.update()

    def rpb_setLineStyle(self, style):
        """
        Line Style setting.
        ...

        Parameters
        --------------

        style: String
            Line style: 'DotLine', 'DashLine', 'SolidLine', passed as a string.

        Raises
        --------------
        none
        """

        if style == 'SolidLine':
            self.rpb_lineStyle = self.lineStyleFlags.SolidLine
        elif style == 'DotLine':
            self.rpb_lineStyle = self.lineStyleFlags.DotLine
        elif style == 'DashLine':
            self.rpb_lineStyle = self.lineStyleFlags.DashLine
        else:
            self.rpb_lineStyle = self.lineStyleFlags.SolidLine

    def rpb_setLineCap(self, cap):
        """
        Line Cap setting.
        ...

        Parameters
        --------------

        cap: String
            Cap is the end point of a stroke. It can be: 'RoundCap' or 'SquareCap'

        Raises
        --------------
        none
        """

        if cap == 'SquareCap':
            self.rpb_lineCap = self.lineCapFlags.SquareCap
        elif cap == 'RoundCap':
            self.rpb_lineCap = self.lineCapFlags.RoundCap

    def rpb_setTextColor(self, rgb):
        """
        Text color of the text inside the progress bar
        ...

        Parameters
        --------------

        rgb: tuple
            Color of the text in the format: (R, G, B)

        Raises
        --------------
        none
        """

        if self.rpb_textColor != rgb:
            self.rpb_textColor = rgb
            self.update()

    def rpb_setTextFont(self, font):
        """
        Font of the text inside the round progress bar
        ...

        Parameters
        --------------

        font: str
            Name of the font in string

        Raises
        --------------
        none
        """

        if self.rpb_textFont != font:
            self.rpb_textFont = font
            self.update()

    def rpb_setTextFormat(self, textTyp):
        """
        Text formatter i.e. the value or the percentage.
        ...

        Parameters
        --------------

        textTyp: str
            'value', 'percentage'

        Raises
        --------------
        none
        """

        if textTyp == 'Value':
            self.rpb_textType = self.textFlags.Value
        elif textTyp == 'Percentage':
            self.rpb_textType = self.textFlags.Percentage
        else:
            self.rpb_textType = self.textFlags.Percentage

    def rpb_setTextRatio(self, ratio):
        """
        Text ratio with respect to the size of the progress bar.
        ...

        Parameters
        --------------

        ratio: int
            In number from 3 to 50 corresponding to 1/3 or 1/50 the size of the RoundProgressBar.

        Raises
        --------------
        none
        """

        if self.rpb_textRatio != ratio:
            if ratio < 3:
                ratio = 3
            elif ratio > 50:
                ratio = 50
            self.rpb_textRatio = ratio
            self.update()

    def rpb_setTextWidth(self, width):
        """
        Text Width.
        ...

        Parameters
        --------------

        font: int
            Text constant width. Will not change during the widget resize.

        Raises
        --------------
        none
        """

        self.dynamicText = False
        if width > 0:
            self.rpb_textWidth = width
            self.update()

    def rpb_setCircleColor(self, rgb):
        """
        Circle color fill inside the circle.
        ...

        Parameters
        --------------

        font: tuple
            The color of the circle in the tuple corresponding to the (R, G, B).

        Raises
        --------------
        none
        """

        if self.rpb_circleColor != rgb:
            self.rpb_circleColor = rgb
            self.update()

    def rpb_setCircleRatio(self, ratio):
        """
        Circle ration corresponding to the round progress bar.
        ...

        Parameters
        --------------

        font: int
            Integer corresponding to the size of the progress bar to that of the round progress bar.

        Raises
        --------------
        none
        """

        if self.rpb_circleRatio != ratio:
            self.rpb_circleRatio = ratio
            self.update()

    def rpb_setPieColor(self, rgb):
        """
        Pie color inside the fill.
        ...

        Parameters
        --------------

        font: tuple
            Tuple consist in format (R, G, B). Same as color setting to Line.

        Raises
        --------------
        none
        """

        if self.rpb_pieColor != rgb:
            self.rpb_pieColor = rgb
            self.update()

    def rpb_setPieRatio(self, ratio):
        """
        Pie Ratio
        ...

        Parameters
        --------------

        font: int
            Ratio corresponding to the size between the RoundProgressBar and the pie size.

        Raises
        --------------
        none
        """

        if self.rpb_pieRatio != ratio:
            self.rpb_pieRatio = ratio
            self.update()

    def rpb_enableText(self, enable):
        """
        Makes the Text visible/Hidden
        ...

        Parameters
        --------------

        font: bool
            True: Text visible, False: Text invisible.

        Raises
        --------------
        none
        """

        if enable:
            self.rpb_textActive = enable
        else:
            self.rpb_textActive = enable
        self.update()

    # ------------------------------------------------------METHODS FOR GETTING THE PROPERTY OF RoundProgressBar SLOTS

    def rpb_getSize(self):
        """
        Get the present size of the progress bar.
        ...

        Returns
        --------------
        Return the size of the round progress bar in int.
        """

        return self.rpb_Size

    def rpb_getValue(self):
        """
        Present value of the progress bar.
        ...

        Returns
        --------------
        int corresponding to the present progress bar value.
        """

        return self.rpb_value / 16

    def rpb_getRange(self):
        """
        Progress bar range.
        ...

        Returns
        --------------
        tuple consisting of minimu and maximum as elements.
        """

        return (self.rpb_minimum, self.rpb_maximum)

    def rpb_getTextWidth(self):
        """
        Text width of the present text in the central of the widget.
        ...

        Returns
        --------------
        int corresponding to the width of the text
        """

        return self.rpb_textWidth

    # ------------------------------------------------------ENGINE: WHERE ALL THE REAL STUFF TAKE PLACE: WORKING OF THE ROUNDPROGRESSBA

    def rpb_MinimumSize(self, dynamicMax, minimum, maximum):
        """
        Minimum size calculating code: Takes consideration of the width of the line/path/circle/pie and the user defined
        width and also the size of the frame/window of the application.

        """

        rpb_Height = self.height()
        rpb_Width = self.width()
        if dynamicMax:
            if rpb_Width >= rpb_Height and rpb_Height >= minimum[1]:
                self.rpb_Size = rpb_Height
            elif rpb_Width < rpb_Height and rpb_Width >= minimum[0]:
                self.rpb_Size = rpb_Width
        else:
            if rpb_Width >= rpb_Height and rpb_Height <= maximum[1]:
                self.rpb_Size = rpb_Height
            elif rpb_Width < rpb_Height and rpb_Width <= maximum[0]:
                self.rpb_Size = rpb_Width

    def convertInputValue(self, value):
        """
        CONVERTS ANY INPUT VALUE TO THE 0*16-360*16 DEGREE REFERENCE OF THE QPainter.drawArc NEEDED.

        """

        self.rpb_value = ((value - self.rpb_minimum) / (self.rpb_maximum - self.rpb_minimum)) * 360 * 16
        self.rpb_value = self.rpb_direction * self.rpb_value
        if self.rpb_textType == RoundProgressBar.textFlags.Percentage:
            self.rpb_textValue = str(
                round(((value - self.rpb_minimum) / (self.rpb_maximum - self.rpb_minimum)) * 100)) + "%"
        else:
            self.rpb_textValue = str(value)

    # SINCE THE THICKNESS OF THE LINE OR THE PATH CAUSES THE WIDGET TO WRONGLY FIT INSIDE THE SIZE OF THE WIDGET DESIGNED IN THE
    # QTDESIGNER, THE CORRECTION FACTOR IS NECESSERY CALLED THE GEOMETRYFACTOR, WHICH CALCULATE THE TWO FACTORS CALLED THE
    # self.posFactor AND THE self.sizeFactor, CALCULATION THIS IS NECESSERY AS THE
    def geometryFactor(self):
        if self.lineWidth > self.pathWidth:
            self.posFactor = self.lineWidth / 2 + 1
            self.sizeFactor = self.lineWidth + 1
        else:
            self.posFactor = self.pathWidth / 2 + 1
            self.sizeFactor = self.pathWidth + 1

    def rpb_textFactor(self):
        if self.dynamicText:
            self.rpb_textWidth = self.rpb_Size / self.rpb_textRatio
        self.textFactorX = self.posFactor + (self.rpb_Size - self.sizeFactor) / 2 - self.rpb_textWidth * 0.75 * (
                    len(self.rpb_textValue) / 2)
        self.textFactorY = self.rpb_textWidth / 2 + self.rpb_Size / 2

    def rpb_circleFactor(self):
        self.rpb_circlePosX = self.positionX + self.posFactor + ((self.rpb_Size) * (1 - self.rpb_circleRatio)) / 2
        self.rpb_circlePosY = self.positionY + self.posFactor + ((self.rpb_Size) * (1 - self.rpb_circleRatio)) / 2

    def rpb_pieFactor(self):
        self.rpb_piePosX = self.positionX + self.posFactor + ((self.rpb_Size) * (1 - self.rpb_pieRatio)) / 2
        self.rpb_piePosY = self.positionY + self.posFactor + ((self.rpb_Size) * (1 - self.rpb_pieRatio)) / 2

    def paintEvent(self, event: QPaintEvent):

        # THIS BELOW CODE AMKE SURE THAT THE SIZE OF THE RoundProgressBar DOESNOT REDUCES TO ZERO WHEN THE USER RESIZES THE WINDOW
        if self.rpb_dynamicMin:
            self.setMinimumSize(QSize(self.lineWidth * 6 + self.pathWidth * 6, self.lineWidth * 6 + self.pathWidth * 6))

        RoundProgressBar.rpb_MinimumSize(self, self.rpb_dynamicMax, self.rpb_minimumSize, self.rpb_maximumSize)
        RoundProgressBar.geometryFactor(self)
        RoundProgressBar.rpb_textFactor(self)
        RoundProgressBar.rpb_circleFactor(self)
        RoundProgressBar.rpb_pieFactor(self)

        if self.rpb_type == 0:  # DONET TYPE
            RoundProgressBar.pathComponent(self)
            RoundProgressBar.lineComponent(self)
            RoundProgressBar.textComponent(self)
        elif self.rpb_type == 1:  # LINE TYPE
            RoundProgressBar.lineComponent(self)
            RoundProgressBar.textComponent(self)
        elif self.rpb_type == 2:  # Pie
            RoundProgressBar.pieComponent(self)
            RoundProgressBar.textComponent(self)
        elif self.rpb_type == 3:  # PIZZA
            RoundProgressBar.circleComponent(self)
            RoundProgressBar.lineComponent(self)
            RoundProgressBar.textComponent(self)
        elif self.rpb_type == 4:  # HYBRID1
            RoundProgressBar.circleComponent(self)
            RoundProgressBar.pathComponent(self)
            RoundProgressBar.lineComponent(self)
            RoundProgressBar.textComponent(self)
        elif self.rpb_type == 5:  # HYBRID2
            RoundProgressBar.pieComponent(self)
            RoundProgressBar.lineComponent(self)
            RoundProgressBar.textComponent(self)

    def lineComponent(self):
        linePainter = QPainter(self)
        linePainter.setRenderHint(QPainter.Antialiasing)
        penLine = QPen()
        penLine.setStyle(self.rpb_lineStyle)
        penLine.setWidth(self.lineWidth)
        penLine.setBrush(QColor(self.lineColor))
        penLine.setCapStyle(self.rpb_lineCap)
        penLine.setJoinStyle(Qt.RoundJoin)
        linePainter.setPen(penLine)
        linePainter.drawArc(self.positionX + self.posFactor, self.positionY + self.posFactor,
                            self.rpb_Size - self.sizeFactor, self.rpb_Size - self.sizeFactor, self.startPosition,
                            self.rpb_value)
        linePainter.end()

    def pathComponent(self):
        pathPainter = QPainter(self)
        pathPainter.setRenderHint(QPainter.Antialiasing)
        penPath = QPen()
        penPath.setStyle(Qt.SolidLine)
        penPath.setWidth(self.pathWidth)
        penPath.setBrush(QColor(self.pathColor))
        penPath.setCapStyle(Qt.RoundCap)
        penPath.setJoinStyle(Qt.RoundJoin)
        pathPainter.setPen(penPath)
        pathPainter.drawArc(self.positionX + self.posFactor, self.positionY + self.posFactor,
                            self.rpb_Size - self.sizeFactor, self.rpb_Size - self.sizeFactor, 0, 360 * 16)
        pathPainter.end()

    def textComponent(self):
        if self.rpb_textActive:
            textPainter = QPainter(self)
            penText = QPen()
            penText.setColor(QColor(self.rpb_textColor))
            textPainter.setPen(penText)
            fontText = QFont()
            fontText.setFamily(self.rpb_textFont)
            fontText.setPointSize(self.rpb_textWidth)
            textPainter.setFont(fontText)
            textPainter.drawText(self.positionX + self.textFactorX, self.positionY + self.textFactorY,
                                 self.rpb_textValue)
            # textPainter.drawText(self.positionX + self.textFactorX, self.positionY + self.textFactorY,
            #                      self.rpb_textStatus)
            textPainter.end()

    def circleComponent(self):
        circlePainter = QPainter(self)
        penCircle = QPen()
        penCircle.setWidth(0)
        penCircle.setColor(QColor(self.rpb_circleColor))
        circlePainter.setRenderHint(QPainter.Antialiasing)
        circlePainter.setPen(penCircle)
        circlePainter.setBrush(QColor(self.rpb_circleColor))
        circlePainter.drawEllipse(self.rpb_circlePosX, self.rpb_circlePosY,
                                  (self.rpb_Size - self.sizeFactor) * self.rpb_circleRatio,
                                  (self.rpb_Size - self.sizeFactor) * self.rpb_circleRatio)

    def pieComponent(self):
        piePainter = QPainter(self)
        penPie = QPen()
        penPie.setWidth(0)
        penPie.setColor(QColor(self.rpb_pieColor))
        piePainter.setRenderHint(QPainter.Antialiasing)
        piePainter.setPen(penPie)
        piePainter.setBrush(QColor(self.rpb_pieColor))
        piePainter.drawPie(self.rpb_piePosX, self.rpb_piePosY, (self.rpb_Size - self.sizeFactor) * self.rpb_pieRatio,
                           (self.rpb_Size - self.sizeFactor) * self.rpb_pieRatio, self.startPosition, self.rpb_value)


# ------------------------------------------------------

x = 0
p = 1


class MyWidget(QtWidgets.QWidget):
    def __init__(self):
        QtWidgets.QWidget.__init__(self)

        self.hello = 'Spiral Progress Bar'
        self.button = QtWidgets.QPushButton("Click me to change Value")
        self.text = QtWidgets.QLabel("Spiral Progress Bar")
        self.text.setAlignment(QtCore.Qt.AlignCenter)

        # CREATING THE SPIRAL PROGRESS BAR OBJECT
        self.spb = RoundProgressBar()

        # ADDING WIDGETS TO THE VERTICAL LAYOUT
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.text)
        self.layout.addWidget(self.button)

        # ADDING THE SPIRAL PROGRESS BAR OBJECT TO THE LAYOUT
        self.layout.addWidget(self.spb)

        self.setLayout(self.layout)
        self.button.clicked.connect(self.magic)  # BUTTON PRESSED EVENT

    def magic(self):
        global x, p
        x = x + 10 * p
        if x == 100:
            p = -1
        elif x == 0:
            p = 1

        # CHANGING THE VALUE OF THE 3 DEFAULT PROGRESS BAR
        self.spb.rpb_setValue((x * 2))

        out_text = 'Spiral Progress Bar: '
        out_text = out_text + str(x) + '%, ' + str(2 * x) + '%, ' + str(3 * x) + '%'
        self.text.setText(out_text)


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    widget = MyWidget()
    widget.show()
    sys.exit(app.exec_())

